"""
Tagulous test: Form TagField

Modules tested:
    tagulous.models.fields.TagField
    tagulous.forms.TagField
"""
from django import forms
from django.test import TestCase

from tagulous import forms as tag_forms
from tagulous import models as tag_models
from tagulous import settings as tag_settings
from tests.lib import TagTestManager, skip_if_mysql
from tests.tagulous_tests_app import forms as test_forms
from tests.tagulous_tests_app import models as test_models


# ##############################################################################
# ###### Test form TagField
# ##############################################################################


class FormTagFieldTest(TagTestManager, TestCase):
    """
    Test form TagFields
    """

    def test_required(self):
        "Test required status is passed from field to widget"
        self.assertTrue(tag_forms.TagField(required=True).required)
        self.assertTrue(tag_forms.TagField(required=True).widget.is_required)
        self.assertFalse(tag_forms.TagField(required=False).required)
        self.assertFalse(tag_forms.TagField(required=False).widget.is_required)
        self.assertTrue(tag_forms.TagField().required)
        self.assertTrue(tag_forms.TagField().widget.is_required)

    def test_input(self):
        "Test valid and invalid input"
        self.assertFieldOutput(
            tag_forms.TagField,
            valid={
                "red": ["red"],
                "Red, Blue": ["Blue", "Red"],
                '"red, blue", yellow': ["red, blue", "yellow"],
            },
            invalid={},
            empty_value=[],
        )

    def test_force_lowercase(self):
        "Test force_lowercase flag set in constructor"
        self.assertFieldOutput(
            tag_forms.TagField,
            field_kwargs={"tag_options": tag_models.TagOptions(force_lowercase=True)},
            valid={
                "red": ["red"],
                "Red, Blue": ["blue", "red"],
                '"Red, Blue", Yellow': ["red, blue", "yellow"],
            },
            invalid={},
            empty_value=[],
        )

    def test_response(self):
        "Test response is as expected"
        form = test_forms.TagFieldForm(data={"tags": "red, blue"})
        self.assertTrue(form.is_valid())
        self.assertEqual(form.cleaned_data["tags"], ["blue", "red"])

    def test_render_basics(self):
        "Check widget renders default settings (not required)"

        class LocalTestForm(forms.Form):
            tag = tag_forms.TagField()

        form = LocalTestForm()
        self.assertHTMLEqual(
            str(form["tag"]),
            (
                '<input autocomplete="off" '
                'data-tag-options="{&quot;required&quot;: true}" '
                'data-tagulous="true" '
                'id="id_tag" name="tag" {{required}}type="text" />'
            ),
        )

    def test_render_tag_optional(self):
        "Check widget renders correctly when field not required"

        class LocalTestForm(forms.Form):
            tag = tag_forms.TagField(required=False)

        form = LocalTestForm()
        self.assertHTMLEqual(
            str(form["tag"]),
            (
                '<input autocomplete="off" '
                'data-tag-options="{&quot;required&quot;: false}" '
                'data-tagulous="true" '
                'id="id_tag" name="tag" type="text" />'
            ),
        )

    def test_render_tag_list(self):
        "Check widget renders data-tag-list"

        class LocalTestForm(forms.Form):
            tag = tag_forms.TagField(autocomplete_tags=["one", "two", "three"])

        form = LocalTestForm()
        self.assertHTMLEqual(
            str(form["tag"]),
            (
                '<input autocomplete="off" '
                'data-tag-options="{&quot;required&quot;: true}" '
                'data-tagulous="true" '
                'data-tag-list="'
                '[&quot;one&quot;, &quot;two&quot;, &quot;three&quot;]" '
                'id="id_tag" name="tag" {{required}}type="text" />'
            ),
        )

    def test_render_tag_url(self):
        "Check widget renders data-tag-url"
        autocomplete_view = "tagulous_tests_app-unlimited"

        class LocalTestForm(forms.Form):
            tag = tag_forms.TagField(
                tag_options=tag_models.TagOptions(autocomplete_view=autocomplete_view)
            )

        form = LocalTestForm()
        self.assertHTMLEqual(
            str(form["tag"]),
            (
                '<input autocomplete="off" '
                'data-tag-options="{&quot;required&quot;: true}" '
                'data-tagulous="true" '
                'data-tag-url="'
                '/tagulous_tests_app/autocomplete/unlimited/" '
                'id="id_tag" name="tag" {{required}}type="text" />'
            ),
        )

    def test_render_value(self):
        "Check widget renders value"
        form = test_forms.TagFieldForm(data={"tags": "run, walk"})
        self.assertHTMLEqual(
            str(form["tags"]),
            (
                '<input autocomplete="off" '
                'data-tag-options="{&quot;required&quot;: true}" '
                'data-tagulous="true" '
                'id="id_tags" name="tags" {{required}}type="text" '
                'value="run, walk" />'
            ),
        )

    def test_render_invalid_tag_url(self):
        "Check widget renders data-tag-url"
        autocomplete_view = "tagulous_tests_app-view_does_not_exist"

        class LocalTestForm(forms.Form):
            tag = tag_forms.TagField(
                tag_options=tag_models.TagOptions(autocomplete_view=autocomplete_view)
            )

        form = LocalTestForm()
        with self.assertRaises(ValueError) as cm:
            str(form["tag"])
        self.assertTrue(
            str(cm.exception).startswith(
                "Invalid autocomplete view: Reverse for '{0}' not found. "
                "'{0}' is not a valid view function or pattern name.".format(
                    autocomplete_view
                )
            )
        )

    def test_render_autocomplete_settings(self):
        "Check widget merges autocomplete settings with defaults"
        # Make a form with some autocomplete settings
        class LocalTestForm(forms.Form):
            tags = tag_forms.TagField(
                tag_options=tag_models.TagOptions(
                    autocomplete_settings={"cats": "purr", "cows": "moo"}
                )
            )

        form = LocalTestForm()

        # Set some defaults in the widget
        self.assertEqual(form["tags"].field.widget.default_autocomplete_settings, None)
        form["tags"].field.widget.default_autocomplete_settings = {
            "bees": "buzz",
            "cats": "mew",
        }

        # Render
        # Expecting bees:buzz, cats:purr, cows:moo
        self.assertHTMLEqual(
            str(form["tags"]),
            (
                '<input autocomplete="off" '
                'data-tag-options="{'
                "&quot;autocomplete_settings&quot;: {&quot;cows&quot;: &quot;moo"
                "&quot;, &quot;bees&quot;: &quot;buzz&quot;, &quot;cats&quot;: "
                '&quot;purr&quot;}, &quot;required&quot;: true}" '
                'data-tagulous="true" '
                'id="id_tags" name="tags" {{required}}type="text" />'
            ),
        )

    def test_invalid_prepare_value(self):
        "Check form field raises exception when given an invalid value"
        form = test_forms.TagFieldForm()
        with self.assertRaises(ValueError) as cm:
            form["tags"].field.prepare_value(["tag", "list", "fails"])
        self.assertEqual(
            str(cm.exception), "Tag field could not prepare unexpected value"
        )


# ##############################################################################
# ###### Test ModelForm TagField
# ##############################################################################


class ModelFormTagFieldTest(TagTestManager, TestCase):
    """
    Test form TagFields
    """

    manage_models = [test_models.TagFieldModel]

    def setUpExtra(self):
        # Load initial tags for all models which have them
        self.form = test_forms.TagFieldModelForm
        self.model = test_models.TagFieldModel
        self.tag_model = self.model.tags.tag_model

    def test_formfield(self):
        "Test that model.TagField.formfield works correctly"
        # Check that the model fields are generated correctly
        tag1_field = self.model._meta.get_field("tags").formfield()
        self.assertIsInstance(tag1_field, tag_forms.TagField)
        self.assertIsInstance(tag1_field.tag_options, tag_models.TagOptions)

        # Check we can get it from the descriptor shortcut method
        tag2_field = self.model.tags.formfield()
        self.assertIsInstance(tag2_field, tag_forms.TagField)
        self.assertIsInstance(tag2_field.tag_options, tag_models.TagOptions)

    def test_media(self):
        "Test that a model form with a TagField has the correct media"
        media = self.form().media
        for js in tag_settings.AUTOCOMPLETE_JS:
            self.assertTrue(js in media._js)
        for grp, files in tag_settings.AUTOCOMPLETE_CSS.items():
            self.assertTrue(grp in media._css)
            for css in files:
                self.assertTrue(css in media._css[grp])

    def test_model_form_save(self):
        """
        Test that a model form with a TagField saves correctly
        """
        form = test_forms.TagFieldModelForm(
            data={"name": "Test 1", "tags": "blue, red"}
        )
        self.assertTrue(form.is_valid())
        t1 = form.save()

        # Check in-memory instance
        self.assertEqual(t1.name, "Test 1")
        self.assertEqual(t1.tags, "blue, red")

        # Check database
        self.assertInstanceEqual(t1, name="Test 1", tags="blue, red")
        self.assertTagModel(self.tag_model, {"blue": 1, "red": 1})

    def test_model_form_save_commit_false(self):
        """
        Test that a model form with a TagField saves correctly when
        commit=False
        """
        form = test_forms.TagFieldModelForm(
            data={"name": "Test 1", "tags": "blue, red"}
        )
        self.assertTrue(form.is_valid())
        t1 = form.save(commit=False)
        t1.save()

        # Check in-memory instance
        self.assertEqual(t1.name, "Test 1")
        self.assertEqual(t1.tags, "")
        self.assertTagModel(self.tag_model, {})
        self.assertInstanceEqual(t1, name="Test 1", tags="")

        # Save M2M data
        form.save_m2m()

        # Check in-memory instance
        self.assertEqual(t1.name, "Test 1")
        self.assertEqual(t1.tags, "blue, red")

        # Check database
        self.assertInstanceEqual(t1, name="Test 1", tags="blue, red")
        self.assertTagModel(self.tag_model, {"blue": 1, "red": 1})

    def test_override_option_dict_formfield(self):
        "Test defaults can be overriden in formfield() call with a dict"
        # Confirm default
        field1 = self.form().fields["tags"]
        self.assertEqual(field1.tag_options.case_sensitive, False)

        # Change default
        field2 = self.model.tags.formfield(tag_options={"case_sensitive": True})
        self.assertEqual(field2.tag_options.case_sensitive, True)

    def test_override_option_cls_formfield(self):
        "Test defaults can be overriden in formfield() call with a TagOption"
        # Confirm default
        field1 = self.form().fields["tags"]
        self.assertEqual(field1.tag_options.case_sensitive, False)

        # Change default
        field2 = self.model.tags.formfield(
            tag_options=tag_models.TagOptions(case_sensitive=True)
        )
        self.assertEqual(field2.tag_options.case_sensitive, True)

    def test_override_autocomplete_tags_formfield(self):
        "Test list of autocomplete tags can be passed in formfield"
        self.tag_model.objects.create(name="red")
        self.tag_model.objects.create(name="blue")
        self.tag_model.objects.create(name="green")

        # Confirm default
        field1 = self.form().fields["tags"]
        self.assertSequenceEqual(
            [t.name for t in field1.autocomplete_tags],
            [t.name for t in self.tag_model.objects.all()],
        )

        # Change default
        field2 = self.model.tags.formfield(autocomplete_tags=["pink", "lime"])
        self.assertSequenceEqual(field2.autocomplete_tags, ["pink", "lime"])

    def test_render_tag_list(self):
        "Check widget renders data-tag-list"
        self.tag_model.objects.create(name="red")
        self.tag_model.objects.create(name="blue")
        self.tag_model.objects.create(name="yellow")
        self.assertTagModel(self.tag_model, {"red": 0, "blue": 0, "yellow": 0})
        form = self.form(data={"name": "Test 1", "tags": "red, blue"})
        self.assertHTMLEqual(
            str(form["tags"]),
            (
                '<input autocomplete="off" '
                'data-tag-options="{&quot;required&quot;: true}" '
                'data-tagulous="true" '
                'data-tag-list="'
                '[&quot;blue&quot;, &quot;red&quot;, &quot;yellow&quot;]" '
                'id="id_tags" name="tags" {{required}}type="text" '
                'value="red, blue" />'
            ),
        )

    def test_initial_string(self):
        "Check initial tag string"
        form = test_forms.TagFieldForm(initial={"tags": "red, blue"})
        self.assertHTMLEqual(
            str(form["tags"]),
            (
                '<input autocomplete="off" '
                'data-tag-options="{&quot;required&quot;: true}" '
                'data-tagulous="true" '
                'id="id_tags" name="tags" {{required}}type="text" '
                'value="red, blue" />'
            ),
        )

    def test_initial_tag_list(self):
        "Check initial tag list"
        t1 = self.tag_model.objects.create(name="red")
        t2 = self.tag_model.objects.create(name="blue")
        form = test_forms.TagFieldForm(initial={"tags": [t1, t2]})
        self.assertHTMLEqual(
            str(form["tags"]),
            (
                '<input autocomplete="off" '
                'data-tag-options="{&quot;required&quot;: true}" '
                'data-tagulous="true" '
                'id="id_tags" name="tags" {{required}}type="text" '
                'value="blue, red" />'
            ),
        )

    def test_initial_tag_queryset(self):
        "Check initial tag queryset"
        self.tag_model.objects.create(name="red")
        self.tag_model.objects.create(name="blue")
        tags = self.tag_model.objects.all()
        form = test_forms.TagFieldForm(initial={"tags": tags})
        self.assertHTMLEqual(
            str(form["tags"]),
            (
                '<input autocomplete="off" '
                'data-tag-options="{&quot;required&quot;: true}" '
                'data-tagulous="true" '
                'id="id_tags" name="tags" {{required}}type="text" '
                'value="blue, red" />'
            ),
        )

    def test_tagged_edit(self):
        "Check edit tagged model form instance works"
        t1 = self.model.objects.create(name="Test 1", tags="blue, red")

        form = self.form(instance=t1)

        self.assertHTMLEqual(
            str(form["tags"]),
            (
                '<input autocomplete="off" '
                'data-tag-list="[&quot;blue&quot;, &quot;red&quot;]" '
                'data-tag-options="{&quot;required&quot;: true}" '
                'data-tagulous="true" '
                'id="id_tags" name="tags" {{required}}type="text" '
                'value="blue, red" />'
            ),
        )

    def test_tagmeta_without_autocomplete_settings(self):
        """
        Check that a tag widget copes with a tag field which takes its options
        from a TagModel with a TagMeta, but where the TagMeta is missing
        autocomplete_settings, and the widget has default_autocomplete_settings.
        """

        class TagMetaUserForm(forms.ModelForm):
            class Meta:
                model = test_models.TagMetaUser
                exclude = []

        form = TagMetaUserForm()
        form["two"].field.widget.default_autocomplete_settings = {"bees": "buzz"}

        self.assertHTMLEqual(
            str(form["two"]),
            (
                '<input autocomplete="off" '
                'data-tag-list="[]" '
                'data-tag-options="{&quot;autocomplete_settings&quot;: '
                "{&quot;bees&quot;: &quot;buzz&quot;}, "
                "&quot;case_sensitive&quot;: true, "
                "&quot;force_lowercase&quot;: true, &quot;max_count&quot;: 10, "
                '&quot;required&quot;: false}" '
                'data-tagulous="true" '
                'id="id_two" name="two" type="text" />'
            ),
        )


# ##############################################################################
# ######  Test TagField blank
# ##############################################################################


class ModelFormTagFieldOptionalTest(TagTestManager, TestCase):
    "Test optional TagField"
    manage_models = [test_models.TagFieldOptionalModel]

    def setUpExtra(self):
        self.form = test_forms.TagFieldOptionalModelForm
        self.model = test_models.TagFieldOptionalModel
        self.tag_model = self.model.tag.tag_model

    def test_optional_tagfield(self):
        "Check model with optional tagfield can be saved when empty"
        form = self.form(data={"name": "Test 1"})
        self.assertTrue(form.is_valid())
        t1 = form.save()
        self.assertEqual(t1.name, "Test 1")
        self.assertEqual(t1.tag.count(), 0)
        self.assertTagModel(self.tag_model, {})


class ModelFormTagFieldRequiredTest(TagTestManager, TestCase):
    "Test required TagField"
    manage_models = [test_models.TagFieldRequiredModel]

    def setUpExtra(self):
        self.form = test_forms.TagFieldRequiredModelForm
        self.model = test_models.TagFieldRequiredModel
        self.tag_model = self.model.tag.tag_model

    def test_required_tagfield_set(self):
        "Check model with required tagfield can be saved when set"
        form = self.form(data={"name": "Test 1", "tag": "red, blue"})
        self.assertTrue(form.is_valid())
        t1 = form.save()
        self.assertEqual(t1.name, "Test 1")
        self.assertEqual(t1.tag, "blue, red")
        self.assertEqual(str(t1.tag), "blue, red")
        self.assertIsInstance(t1.tag, tag_models.TagRelatedManagerMixin)
        self.assertEqual(t1.tag.count(), 2)
        self.assertEqual(t1.tag.all()[0], "blue")
        self.assertEqual(t1.tag.all()[1], "red")
        self.assertTagModel(self.tag_model, {"red": 1, "blue": 1})

    def test_required_tagfield_empty(self):
        "Check model with required tagfield cannot be saved when empty"
        form = self.form(data={"name": "Test 1"})
        self.assertFalse(form.is_valid())
        self.assertEqual(len(form.errors.keys()), 1)
        self.assertEqual(list(form.errors.keys())[0], "tag")
        self.assertEqual(len(form.errors["tag"]), 1)
        self.assertEqual(form.errors["tag"][0], "This field is required.")


# ##############################################################################
# ###### Test form TagField options
# ##############################################################################


class ModelFormTagFieldOptionsTest(TagTestManager, TestCase):
    manage_models = [test_models.TagFieldOptionsModel]

    def setUpExtra(self):
        # Load initial tags for all models which have them
        self.form = test_forms.TagFieldOptionsModelForm
        self.model = test_models.TagFieldOptionsModel

    @skip_if_mysql
    def test_case_sensitive_true(self):
        "Test form TagField case_sensitive true"
        # Prep tag model
        tag_model = self.model.case_sensitive_true.tag_model
        self.assertTagModel(tag_model, {"Adam": 0})

        # Check that the option is passed to the form field from model field
        form = self.form(data={"case_sensitive_true": "adam"})
        field = form.fields["case_sensitive_true"]
        self.assertIsInstance(field, tag_forms.TagField)
        self.assertEqual(field.tag_options.case_sensitive, True)

        # Check that the option is passed to the widget
        self.assertEqual(field.widget.tag_options.case_sensitive, True)

        # No effect on form data
        self.assertTrue(form.is_valid())
        self.assertEqual(form.cleaned_data["case_sensitive_true"], ["adam"])

        # No effect on save
        obj = form.save()
        self.assertEqual(obj.case_sensitive_true, "adam")
        self.assertTagModel(tag_model, {"Adam": 0, "adam": 1})

    @skip_if_mysql
    def test_case_sensitive_false(self):
        "Test form TagField case_sensitive false"
        # Prep tag model
        tag_model = self.model.case_sensitive_false.tag_model

        # Check that the option is passed to the form field from model field
        form = self.form(data={"case_sensitive_false": "adam"})
        field = form.fields["case_sensitive_false"]
        self.assertIsInstance(field, tag_forms.TagField)
        self.assertEqual(field.tag_options.case_sensitive, False)
        self.assertEqual(field.widget.tag_options.case_sensitive, False)

        # Still no effect on form data
        self.assertTrue(form.is_valid())
        self.assertEqual(form.cleaned_data["case_sensitive_false"], ["adam"])

        # But when saved, the instance will be capitalised
        obj = form.save()
        self.assertEqual(obj.case_sensitive_false, "Adam")
        self.assertTagModel(tag_model, {"Adam": 1})

    def test_force_lowercase_true(self):
        "Test form TagField force_lowercase true"
        # Prep tag model
        tag_model = self.model.force_lowercase_true.tag_model

        # Check that the option is passed to the form field from model field
        form = self.form(data={"force_lowercase_true": "Adam"})
        field = form.fields["force_lowercase_true"]
        self.assertIsInstance(field, tag_forms.TagField)
        self.assertEqual(field.tag_options.force_lowercase, True)
        self.assertEqual(field.widget.tag_options.force_lowercase, True)

        # Form data returns lowercase case
        self.assertTrue(form.is_valid())
        self.assertEqual(form.cleaned_data["force_lowercase_true"], ["adam"])

        # When saved, the instance will be lowercase
        obj = form.save()
        self.assertEqual(obj.force_lowercase_true, "adam")
        self.assertTagModel(tag_model, {"adam": 1})

    def test_force_lowercase_false(self):
        "Test form TagField force_lowercase false"
        # Check that the option is passed to the form field from model field
        form = self.form(data={"force_lowercase_false": "Adam"})
        field = form.fields["force_lowercase_false"]
        self.assertIsInstance(field, tag_forms.TagField)
        self.assertEqual(field.tag_options.force_lowercase, False)
        self.assertEqual(field.widget.tag_options.force_lowercase, False)

        # Check the input string is returned in the new case
        self.assertTrue(form.is_valid())
        self.assertEqual(form.cleaned_data["force_lowercase_false"], ["Adam"])

    def test_max_count(self):
        "Test form TagField max_count is passed and enforced"
        # It's not relevant
        # Check that the option is passed to the form field from model field
        form = self.form(data={"max_count": "one"})
        field = form.fields["max_count"]
        self.assertIsInstance(field, tag_forms.TagField)
        self.assertEqual(field.tag_options.max_count, 3)

        # Check the input string is returned in the new case
        self.assertTrue(form.is_valid())
        self.assertEqual(form.cleaned_data["max_count"], ["one"])

        # Check two
        form = self.form(data={"max_count": "one, two"})
        self.assertTrue(form.is_valid())
        self.assertEqual(form.cleaned_data["max_count"], ["one", "two"])

        # Check three
        form = self.form(data={"max_count": "one, two, three"})
        self.assertTrue(form.is_valid())
        self.assertEqual(
            form.cleaned_data["max_count"], sorted(["one", "two", "three"])
        )

        # Check four
        form = self.form(data={"max_count": "one, two, three, four"})
        self.assertFalse(form.is_valid())
        self.assertEqual(len(form.errors.keys()), 1)
        self.assertEqual(list(form.errors.keys())[0], "max_count")
        self.assertEqual(len(form.errors["max_count"]), 1)
        self.assertEqual(
            form.errors["max_count"][0], "This field can only have 3 arguments"
        )

    def text_max_count_1(self):
        "Test form TagField max_count of 1"
        # Mostly just to test grammar of the error message
        class LocalTestForm(forms.Form):
            tags = tag_forms.TagField(tag_options=tag_models.TagOptions(max_count=1))

        form = self.form(data={"max_count": "one"})
        self.assertTrue(form.is_valid())
        self.assertEqual(form.cleaned_data["max_count"], ["one"])

        # Check two
        form = self.form(data={"max_count": "one, two"})
        self.assertFalse(form.is_valid())
        self.assertEqual(len(form.errors.keys()), 1)
        self.assertEqual(list(form.errors.keys())[0], "max_count")
        self.assertEqual(len(form.errors["max_count"]), 1)
        self.assertEqual(
            form.errors["max_count"][0], "This field can only have 1 argument"
        )

    def test_initial_without_autocomplete_initial(self):
        "Check a field with initial but without autocomplete_initial lists all"
        tag_model = self.model.initial_string.tag_model
        tag_model.objects.create(name="David")
        tag_model.objects.create(name="Eric")
        tag_model.objects.create(name="Frank")

        # Confirm default
        field = self.form().fields["initial_string"]
        self.assertSequenceEqual(
            [t.name for t in field.autocomplete_tags],
            [t.name for t in tag_model.objects.all()],
        )

    def test_initial_with_autocomplete_initial(self):
        "Check a field with initial and autocomplete_initial lists initial"
        tag_model = self.model.initial_list.tag_model
        tag_model.objects.create(name="David")
        tag_model.objects.create(name="Eric")
        tag_model.objects.create(name="Frank")

        # Confirm default
        field = self.form().fields["initial_list"]
        self.assertSequenceEqual(
            [t.name for t in field.autocomplete_tags],
            [t.name for t in tag_model.objects.initial()],
        )



"""
Tagulous models

Make important functions and classes available on tagulous.models
"""
from ..signals.pre import register_pre_signals
from . import initial, migrations  # noqa
from .descriptors import BaseTagDescriptor, SingleTagDescriptor, TagDescriptor  # noqa
from .fields import (  # noqa
    BaseTagField,
    SingleTagField,
    TagField,
    singletagfields_from_model,
    tagfields_from_model,
)
from .managers import SingleTagManager, TagRelatedManagerMixin  # noqa
from .models import (  # noqa
    BaseTagModel,
    BaseTagTreeModel,
    TagModel,
    TagModelManager,
    TagModelQuerySet,
    TagTreeModel,
)
from .options import TagOptions  # noqa
from .tagged import TaggedManager, TaggedModel, TaggedQuerySet  # noqa


register_pre_signals()



"""
Tagulous test: Forms with SingleTagField and TagFields

Modules tested:
    tagulous.models.fields.SingleTagField
    tagulous.models.fields.TagField
    tagulous.forms.SingleTagField
    tagulous.forms.TagField
"""
from django.test import TestCase

from tests.lib import TagTestManager
from tests.tagulous_tests_app import forms as test_forms
from tests.tagulous_tests_app import models as test_models


# ##############################################################################
# ###### Test mixed form
# ##############################################################################


class FormMixedNonTagRefTest(TagTestManager, TestCase):
    """
    Test form TagFields
    """

    manage_models = [test_models.MixedNonTagRefTest]

    def setUpExtra(self):
        self.model = test_models.MixedNonTagRefTest
        self.tag_model = self.model.singletag.tag_model

    def test_model_form_save(self):
        """
        Test that a model form with a TagField saves correctly with other
        relationships
        """
        tag1 = self.tag_model.objects.create(name="blue")
        tag2 = self.tag_model.objects.create(name="red")
        tag3 = self.tag_model.objects.create(name="green")
        form = test_forms.MixedNonTagRefModelForm(
            data={
                "name": "Test 1",
                "singletag": "purple",
                "tags": "yellow, orange",
                "fk": tag1.pk,
                "mm": [tag2.pk, tag3.pk],
            }
        )
        self.assertTrue(form.is_valid())
        t1 = form.save()

        # Check in-memory instance
        self.assertEqual(t1.name, "Test 1")
        self.assertEqual(t1.singletag, "purple")
        self.assertEqual(t1.tags, "orange, yellow")
        self.assertEqual(t1.fk, tag1)
        self.assertEqual(list(t1.mm.all().order_by("pk")), [tag2, tag3])

        # Check database
        self.assertInstanceEqual(
            t1,
            name="Test 1",
            singletag="purple",
            tags="orange, yellow",
            fk=tag1,
            mm=[tag2, tag3],
        )
        self.assertTagModel(
            self.tag_model,
            {"blue": 0, "red": 0, "green": 0, "purple": 1, "orange": 1, "yellow": 1},
        )

    def test_model_form_save_commit_false(self):
        """
        Test that a model form with a TagField saves correctly when save_m2m
        is also called
        """
        tag1 = self.tag_model.objects.create(name="blue")
        tag2 = self.tag_model.objects.create(name="red")
        tag3 = self.tag_model.objects.create(name="green")
        form = test_forms.MixedNonTagRefModelForm(
            data={
                "name": "Test 1",
                "singletag": "purple",
                "tags": "orange, yellow",
                "fk": tag1.pk,
                "mm": [tag2.pk, tag3.pk],
            }
        )
        self.assertTrue(form.is_valid())
        t1 = form.save(commit=False)
        t1.save()
        form.save_m2m()

        # Check in-memory instance
        self.assertEqual(t1.name, "Test 1")
        self.assertEqual(t1.singletag, "purple")
        self.assertEqual(t1.tags, "orange, yellow")
        self.assertEqual(t1.fk, tag1)
        self.assertEqual(list(t1.mm.all().order_by("pk")), [tag2, tag3])

        # Check database
        self.assertInstanceEqual(
            t1,
            name="Test 1",
            singletag="purple",
            tags="orange, yellow",
            fk=tag1,
            mm=[tag2, tag3],
        )
        self.assertTagModel(
            self.tag_model,
            {"blue": 0, "red": 0, "green": 0, "purple": 1, "orange": 1, "yellow": 1},
        )


# ##############################################################################
# ###### Test inline formsets
# ##############################################################################


class SingleInlineFormsetTest(TagTestManager, TestCase):
    """
    Test inline formset for SingleTagField relationships
    """

    manage_models = [test_models.SimpleMixedTest]

    def setUpExtra(self):
        self.model = test_models.SimpleMixedTest
        self.tag_model = self.model.singletag.tag_model
        self.formset = test_forms.SimpleMixedSingleInline

    def test_add(self):
        tag1 = self.tag_model.objects.create(name="Mr")
        self.assertTagModel(self.tag_model, {"Mr": 0})
        self.assertEqual(self.model.objects.count(), 0)

        formset = self.formset(
            data={
                "formset-TOTAL_FORMS": "3",
                "formset-INITIAL_FORMS": "0",
                "formset-MAX_NUM_FORMS": "1000",
                "formset-0-name": "Test 1",
                "formset-0-DELETE": "",
                "formset-1-name": "Test 2",
                "formset-1-DELETE": "",
            },
            instance=tag1,
            prefix="formset",
        )
        self.assertTrue(formset.is_valid())
        formset.save()

        self.assertEqual(self.model.objects.count(), 2)
        obj1 = self.model.objects.get(name="Test 1")
        obj2 = self.model.objects.get(name="Test 2")
        self.assertEqual(obj1.singletag, tag1)
        self.assertEqual(obj2.singletag, tag1)
        self.assertTagModel(self.tag_model, {"Mr": 2})

    def test_edit_pk(self):
        "Check that edit formset forms refer to the tag pk, not its name"
        obj1 = self.model.objects.create(name="Test 1", singletag="Mr")
        self.model.objects.create(name="Test 2", singletag="Mr")
        tag1 = obj1.singletag
        self.assertTagModel(self.tag_model, {"Mr": 2})
        self.assertEqual(self.model.objects.count(), 2)

        formset = self.formset(prefix="formset", instance=tag1)
        self.assertHTMLEqual(
            str(formset.forms[0]["singletag"]),
            (
                '<input type="hidden" name="formset-0-singletag" '
                'value="%d" id="id_formset-0-singletag" />'
            )
            % tag1.pk,
        )

    def test_edit(self):
        obj1 = self.model.objects.create(name="Test 1", singletag="Mr")
        obj2 = self.model.objects.create(name="Test 2", singletag="Mr")
        tag1 = obj1.singletag
        self.assertTagModel(self.tag_model, {"Mr": 2})
        self.assertEqual(self.model.objects.count(), 2)

        formset = self.formset(
            data={
                "formset-TOTAL_FORMS": "5",
                "formset-INITIAL_FORMS": "2",
                "formset-MAX_NUM_FORMS": "1000",
                "formset-0-name": "Test 1e",
                "formset-0-id": "%d" % obj1.pk,
                "formset-0-singletag": "%d" % tag1.pk,
                "formset-0-DELETE": "",
                "formset-1-name": "Test 2e",
                "formset-1-id": "%d" % obj2.pk,
                "formset-1-singletag": "%d" % tag1.pk,
                "formset-1-DELETE": "",
            },
            instance=tag1,
            prefix="formset",
        )
        formset.is_valid()
        self.assertTrue(formset.is_valid())
        formset.save()

        self.assertEqual(self.model.objects.count(), 2)
        obj1 = self.model.objects.get(name="Test 1e")
        obj2 = self.model.objects.get(name="Test 2e")
        self.assertEqual(obj1.singletag, tag1)
        self.assertEqual(obj2.singletag, tag1)
        self.assertTagModel(self.tag_model, {"Mr": 2})



"""
Test models
"""
from django.db import models

import tagulous


try:
    import django.apps as django_apps
except ImportError:
    django_apps = None


#
# MigrationTestModel will be created as needed
#

app_name = "tagulous_tests_migration"


def clear_django():
    "Clear app from django's model cache"
    # Models to clear out from loading cache
    model_names = [
        # Test model
        "migrationtestmodel",
        # Tagulous models
        "tagulous_migrationtestmodel_singletag",
        "tagulous_migrationtestmodel_tags",
        # Django through models
        "migrationtestmodel_tags",
    ]

    # Clear the get_models LRU cache
    django_apps.apps.clear_cache()

    # Find dict container for app models cache
    app_config = django_apps.apps.get_app_config(app_name)
    app_models = app_config.models

    # Clear the named models from the app models cache
    for lower_name in model_names:
        try:
            del app_models[lower_name]
        except KeyError:
            pass


def unset_model():
    "Remove the model"
    if "MigrationTestModel" in globals():
        del globals()["MigrationTestModel"]
    clear_django()


def set_model_initial():
    "Create an initial model without tag fields"
    clear_django()
    return type(
        str("MigrationTestModel"),
        (models.Model,),
        {
            "__module__": "tests.tagulous_tests_migration.models",
            "name": models.CharField(max_length=10),
        },
    )


def set_model_tagged():
    "Initial model with tag fields"
    clear_django()
    model = type(
        str("MigrationTestModel"),
        (models.Model,),
        {
            "__module__": "tests.tagulous_tests_migration.models",
            "name": models.CharField(max_length=10),
            "singletag": tagulous.models.SingleTagField(blank=True, null=True),
            "tags": tagulous.models.TagField(),
        },
    )

    # Just confirm dynamic creation worked as expected
    assert issubclass(model, tagulous.models.tagged.TaggedModel), "Model is not tagged"
    assert issubclass(
        model.singletag.tag_model, tagulous.models.models.TagModel
    ), "Single tag model not TagModel"
    assert issubclass(
        model.tags.tag_model, tagulous.models.models.TagModel
    ), "Tag model not TagModel"
    return model


def set_model_tree():
    "Tagged model with tags field as tree"
    clear_django()
    model = type(
        str("MigrationTestModel"),
        (models.Model,),
        {
            "__module__": "tests.tagulous_tests_migration.models",
            "name": models.CharField(max_length=10),
            "singletag": tagulous.models.SingleTagField(blank=True, null=True),
            str("tags"): tagulous.models.TagField(tree=True),
        },
    )

    # Just confirm dynamic creation worked as expected
    assert issubclass(model, tagulous.models.tagged.TaggedModel), "Model is not tagged"
    assert issubclass(
        model.singletag.tag_model, tagulous.models.models.TagModel
    ), "Single tag model not TagModel"
    assert issubclass(
        model.tags.tag_model, tagulous.models.models.TagModel
    ), "Tag model not TagModel"
    return model



"""
XML serializer with Tagulous support
"""
from django.core.serializers import xml_serializer

from ..models.fields import SingleTagField, TagField
from ..models.tagged import TaggedModel
from . import base


class FakeObject(object):
    """
    Fake object to feed into the standard XML serializer, to trick it into
    serializing a tag field as a text field
    """

    def __init__(self, field_name, value):
        setattr(self, field_name, value)
        self.value = value


class FakeField(object):
    """
    Fake field to feed into the standard XML serializer, to trick it into
    serializing a tag field as a text field
    """

    def __init__(self, name):
        self.name = name

    def get_internal_type(self):
        return "TextField"

    def value_to_string(self, obj):
        return obj.value


class Serializer(xml_serializer.Serializer):
    """
    XML serializer with tag field support
    """

    def handle_tagfield(self, obj, field):
        """
        Trick XML serializer into serializing this as text
        """
        tag_string = str(getattr(obj, field.name))
        fake_obj = FakeObject(field.name, tag_string)
        fake_field = FakeField(field.name)
        self.handle_field(fake_obj, fake_field)

    def handle_fk_field(self, obj, field):
        if isinstance(field, SingleTagField):
            self.handle_tagfield(obj, field)
        else:
            super(Serializer, self).handle_fk_field(obj, field)

    def handle_m2m_field(self, obj, field):
        if isinstance(field, TagField):
            self.handle_tagfield(obj, field)
        else:
            super(Serializer, self).handle_m2m_field(obj, field)


class Deserializer(xml_serializer.Deserializer):
    def _handle_object(self, node):
        obj = super(Deserializer, self)._handle_object(node)
        return base._deserialize_obj(obj)

    # Could override _handle_fk_field_node and _handle_m2m_field_node with
    # code to check for tag fields, but the code to turn them into fake
    # textfields is already in place

    def _get_model_from_node(self, node, attr):
        RealModel = super(Deserializer, self)._get_model_from_node(node, attr)

        if issubclass(RealModel, TaggedModel):
            Model = RealModel._detag_to_serializable()
        else:
            Model = RealModel

        return Model



import os
import re
import sys

from setuptools import find_packages, setup


VERSION = "1.2.1"


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


def runtests(args):
    "Run tests"
    import django
    from django.conf import settings
    from django.core.management import execute_from_command_line

    if not settings.configured:
        testenv = re.sub(
            r"[^a-zA-Z0-9]",
            "_",
            os.environ.get("TOXENV", "_".join(str(v) for v in django.VERSION)),
        )
        tests_migration_module_name = "migrations_{}".format(testenv)

        SETTINGS = dict(
            INSTALLED_APPS=[
                "django.contrib.auth",
                "django.contrib.admin",
                "django.contrib.sessions",
                "django.contrib.contenttypes",
                "django.contrib.messages",
                "tagulous",
                "tests",
                "tests.tagulous_tests_app",
                "tests.tagulous_tests_app2",
                "tests.tagulousTestsApp3",
                "tests.tagulous_tests_migration",
            ],
            MIDDLEWARE=[
                "django.middleware.common.CommonMiddleware",
                "django.contrib.sessions.middleware.SessionMiddleware",
                "django.contrib.auth.middleware.AuthenticationMiddleware",
                "django.contrib.messages.middleware.MessageMiddleware",
            ],
            SECRET_KEY="secret",
            DEFAULT_AUTO_FIELD="django.db.models.AutoField",
            ROOT_URLCONF="tests.tagulous_tests_app.urls",
            SERIALIZATION_MODULES={
                "xml": "tagulous.serializers.xml_serializer",
                "json": "tagulous.serializers.json",
                "python": "tagulous.serializers.python",
            },
            TEMPLATES=[
                {
                    "BACKEND": "django.template.backends.django.DjangoTemplates",
                    "APP_DIRS": True,
                    "OPTIONS": {
                        "context_processors": [
                            "django.contrib.auth.context_processors.auth",
                            "django.contrib.messages.context_processors.messages",
                            "django.template.context_processors.request",
                        ],
                    },
                }
            ],
            MIGRATION_MODULES={
                "tagulous_tests_migration": "tests.tagulous_tests_migration.{}".format(
                    tests_migration_module_name
                ),
            },
            TAGULOUS_NAME_MAX_LENGTH=191,
        )

        # If yaml is available, add to serialisers
        try:
            import yaml  # noqa
        except ImportError:
            pass
        else:
            SETTINGS["SERIALIZATION_MODULES"]["yaml"] = "tagulous.serializers.pyyaml"

        # Build database settings
        DATABASE = {
            "ENGINE": "django.db.backends.sqlite3",
            "NAME": ":memory:",
        }
        engine = os.environ.get("DATABASE_ENGINE")
        if engine:
            if engine == "postgresql":
                DATABASE["ENGINE"] = "django.db.backends.postgresql_psycopg2"
                DATABASE["HOST"] = "localhost"
            elif engine == "mysql":
                DATABASE["ENGINE"] = "django.db.backends.mysql"

                # Make sure test DB is going to be UTF8
                DATABASE_TEST = {
                    "TEST": {"CHARSET": "utf8", "COLLATION": "utf8_general_ci"}
                }
                DATABASE.update(DATABASE_TEST)

            else:
                raise ValueError("Unknown database engine")

            DATABASE["NAME"] = os.environ.get(
                "DATABASE_NAME", "test_tagulous_%s" % testenv
            )
            for key in ["USER", "PASSWORD", "HOST", "PORT"]:
                if "DATABASE_" + key in os.environ:
                    DATABASE[key] = os.environ["DATABASE_" + key]
        SETTINGS["DATABASES"] = {"default": DATABASE, "test": DATABASE}

        # Make sure the django migration loader will find MIGRATION_MODULES
        # This will be cleaned away after each test
        tests_migration_path = os.path.join(
            os.path.dirname(__file__), "tests", "tagulous_tests_migration",
        )
        if not os.path.isdir(tests_migration_path):
            raise ValueError("tests.tagulous_tests_migration not found")

        tests_migration_module_path = os.path.join(
            tests_migration_path, tests_migration_module_name,
        )
        if not os.path.exists(tests_migration_module_path):
            os.mkdir(tests_migration_module_path)

        tests_migration_module_init = os.path.join(
            tests_migration_module_path, "__init__.py",
        )
        if not os.path.exists(tests_migration_module_init):
            open(tests_migration_module_init, "a").close()

        # Configure
        settings.configure(**SETTINGS)

    execute_from_command_line(args[:1] + ["test"] + (args[2:] or ["tests"]))


if len(sys.argv) > 1 and sys.argv[1] == "test":
    runtests(sys.argv)
    sys.exit()

setup(
    name="django-tagulous",
    version=VERSION,
    author="Richard Terry",
    author_email="code@radiac.net",
    description=("Fabulous Tagging for Django"),
    license="BSD",
    keywords="django tagging",
    url="http://radiac.net/projects/django-tagulous/",
    long_description=read("README.rst"),
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Web Environment",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Framework :: Django",
        "Framework :: Django :: 2.2",
        "Framework :: Django :: 3.0",
        "Framework :: Django :: 3.1",
        "Framework :: Django :: 3.2",
    ],
    install_requires=["Django>=2.2"],
    extras_require={"dev": ["tox", "jasmine"], "devdb": ["psycopg2", "mysqlclient"],},
    zip_safe=True,
    packages=find_packages(exclude=("docs", "tests*",)),
    include_package_data=True,
)



"""
JSON serializer with Tagulous support
"""
from django.core.serializers import json as json_serializer

from . import base


class Serializer(base.SerializerMixin, json_serializer.Serializer):
    """
    JSON serializer with tag field support
    """

    pass


Deserializer = base.DeserializerWrapper(
    json_serializer.Deserializer,
    doc="Deserialize a stream or string of JSON data, with tag field support",
)



"""
Extensions for serializers to add tag field support
"""
from ..models.fields import SingleTagField, TagField
from ..models.tagged import TaggedModel


class SerializerMixin(object):
    """
    Mixin for any Serializer which doesn't define its own handle_fk_field or
    handle_m2m_field, to add tag support
    """

    def handle_fk_field(self, obj, field):
        if isinstance(field, SingleTagField):
            self._current[field.name] = str(getattr(obj, field.name))
        else:
            super(SerializerMixin, self).handle_fk_field(obj, field)

    def handle_m2m_field(self, obj, field):
        if isinstance(field, TagField):
            self._current[field.name] = [
                tag.name for tag in getattr(obj, field.name).all()
            ]
        else:
            super(SerializerMixin, self).handle_m2m_field(obj, field)


def _deserialize_obj(obj):
    """
    Given a DeserializationObject, deserialise any tag fields
    """
    # Convert any tag fields from m2m to string assignments
    Model = obj.object.__class__
    if hasattr(Model, "_retag_to_original"):
        obj.object = obj.object._retag_to_original()
    return obj


def DeserializerWrapper(deserializer, doc=None):
    """
    Adds tag support to any deserializer which yields DeserializedObjects

    Given a deserializer, it modifies the DeserializedObjects which would
    normally be returned so tag fields are deserialized correctly.
    """

    def wrapper(object_list, **options):
        # Call normal deserializer, get generator of DeserializedObjects
        obj_generator = deserializer(object_list, **options)
        for obj in obj_generator:
            yield _deserialize_obj(obj)

    if doc:
        wrapper.__doc__ = doc
    return wrapper


def monkeypatch_get_model(serializer):
    """
    Given a model identifier, get the model - unless it's a TaggedModel, in
    which case return a temporary fake model to get it serialized.
    """
    old_get_model = serializer._get_model

    def _get_model(model_identifier):
        RealModel = old_get_model(model_identifier)

        if issubclass(RealModel, TaggedModel):
            Model = RealModel._detag_to_serializable()
        else:
            Model = RealModel

        return Model

    serializer._get_model = _get_model



import json

from django import forms
from django.contrib.admin.widgets import AutocompleteMixin
from django.core.serializers.json import DjangoJSONEncoder
from django.db.models.query import QuerySet
from django.urls import NoReverseMatch, reverse
from django.utils.encoding import force_str
from django.utils.html import escape
from django.utils.translation import gettext as _

from . import settings
from .models import options
from .models.models import BaseTagModel, TagModelQuerySet
from .utils import parse_tags, render_tags


class TagWidgetBase(forms.TextInput):
    """
    Base class for tag widgets
    """

    # Attributes that subclasses must set
    autocomplete_settings = None

    # Attributes that the calling Field must set
    tag_options = None
    autocomplete_tags = None

    # Provide choices attribute for admin site, to avoid an error in the event
    # tagulous.admin isn't used to register the admin model
    choices = None

    def render(self, name, value, attrs=None, renderer=None):
        if attrs is None:
            attrs = {}

        # Try to provide a URL for the autocomplete to load tags on demand
        autocomplete_view = self.tag_options.autocomplete_view
        if autocomplete_view:
            try:
                attrs["data-tag-url"] = reverse(
                    autocomplete_view,
                    args=self.tag_options.autocomplete_view_args,
                    kwargs=self.tag_options.autocomplete_view_kwargs,
                )
            except NoReverseMatch as e:
                raise ValueError("Invalid autocomplete view: %s" % e)

        # Otherwise embed them, if provided
        elif self.autocomplete_tags is not None:
            autocomplete_tags = self.autocomplete_tags
            # If it's a queryset, make sure it hasn't been consumed, otherwise
            # changes won't show in the list
            if isinstance(autocomplete_tags, QuerySet):
                autocomplete_tags = autocomplete_tags.all()

            attrs["data-tag-list"] = escape(
                force_str(
                    json.dumps(
                        # Call str rather than tag.name directly, in case
                        # we've been given a list of tag strings
                        [str(tag) for tag in autocomplete_tags],
                        cls=DjangoJSONEncoder,
                    )
                )
            )

        # Merge default autocomplete settings into tag options
        tag_options = self.tag_options.form_items(with_defaults=False)
        if self.default_autocomplete_settings is not None:
            autocomplete_settings = self.default_autocomplete_settings.copy()
            autocomplete_settings.update(tag_options.get("autocomplete_settings", {}))
            tag_options["autocomplete_settings"] = autocomplete_settings

        # Inject extra settings
        tag_options.update({"required": self.is_required})

        # Store tag options
        attrs["data-tag-options"] = escape(
            force_str(json.dumps(tag_options, cls=DjangoJSONEncoder))
        )

        # Mark it for the javascript to find
        attrs["data-tagulous"] = "true"

        # Turn off browser's autocomplete
        attrs["autocomplete"] = "off"

        # Render value
        return super(TagWidgetBase, self).render(name, value, attrs)


class TagWidget(TagWidgetBase):
    """
    Tag widget for public-facing forms
    """

    default_autocomplete_settings = settings.AUTOCOMPLETE_SETTINGS

    class Media:
        css = settings.AUTOCOMPLETE_CSS
        js = settings.AUTOCOMPLETE_JS


class AdminTagWidget(TagWidgetBase):
    """
    Tag widget for admin forms
    """

    default_autocomplete_settings = settings.ADMIN_AUTOCOMPLETE_SETTINGS

    # Admin will be expecting this to have a choices attribute
    # Set this so the admin will behave as expected
    choices = None

    @property
    def media(self):
        # Get the media from the AutocompleteMixin - this will give us Django's
        # vendored jQuery and select2
        media = AutocompleteMixin.media.fget(None)

        return media + forms.Media(
            js=settings.ADMIN_AUTOCOMPLETE_JS, css=settings.ADMIN_AUTOCOMPLETE_CSS,
        )

    def render(self, name, value, attrs=None, renderer=None):
        if attrs is None:
            attrs = {}
        attrs["data-theme"] = "admin-autocomplete"

        return super().render(name, value, attrs, renderer)


class BaseTagField(forms.CharField):
    """
    Base class for form tag fields
    """

    # Use the tag widget
    widget = TagWidget

    def __init__(self, tag_options=None, autocomplete_tags=None, **kwargs):
        """
        Takes all CharField options, plus:
            tag_options         A TagOptions instance
                                If not provided, uses Tagulous defaults
            autocomplete_tags   Iterable of tags to be used for autocomplete,
                                ie a queryset from a TagModel, or a list of
                                strings. Will be ignored if tag_options
                                contains autocomplete_view
        """
        # Initialise as normal
        super(BaseTagField, self).__init__(**kwargs)

        # Add tag options and autocomplete tags
        # Will use getters and setters to mirror onto widget
        self.tag_options = tag_options or options.TagOptions()
        self.autocomplete_tags = autocomplete_tags

    def prepare_value(self, value):
        """
        Prepare the value - a tag string
        """
        # Catch no value (empty form for add)
        if not value:
            tag_string = ""

        # Will normally get a string
        elif isinstance(value, str):
            tag_string = value

        # Otherwise will be given by the model's TagField.value_from_object().
        # The value comes from django.forms.model.model_to_dict, which thinks
        # it produced a list of pks - but TagField.value_from_object tricked it
        # by returning a list with a single item, the tag string.
        else:
            # Catch changes in model_to_dict which broke the trick
            if len(value) != 1:
                raise ValueError(_("Tag field could not prepare unexpected value"))
            tag_string = value[0]

        return super(BaseTagField, self).prepare_value(tag_string)

    # Use setters and getters to ensure any changes to the field are mirrored
    # in the widget, in the same way ModelChoiceField mirrors its queryset
    def _get_tag_options(self):
        return self._tag_options

    def _prepare_tag_options(self, tag_options):
        "Clone tag options"
        return tag_options + options.TagOptions()

    def _set_tag_options(self, tag_options):
        tag_options = self._prepare_tag_options(tag_options)
        self._tag_options = tag_options
        self.widget.tag_options = tag_options

    tag_options = property(_get_tag_options, _set_tag_options)

    def _get_autocomplete_tags(self):
        return self._autocomplete_tags

    def _set_autocomplete_tags(self, autocomplete_tags):
        self._autocomplete_tags = autocomplete_tags
        self.widget.autocomplete_tags = autocomplete_tags

    autocomplete_tags = property(_get_autocomplete_tags, _set_autocomplete_tags)

    def widget_attrs(self, widget):
        """
        Given a Widget instance (*not* a Widget class), returns a dictionary of
        any HTML attributes that should be added to the Widget, based on this
        Field.
        """
        return {}

    def to_python(self, value):
        # Tags shouldn't be whitespace. Strip here for ``required`` validation
        if isinstance(value, str):
            value = value.strip()

        return super(BaseTagField, self).to_python(value)

    def clean(self, value, single=False):
        """
        Clean the form value

        If optional argument ``single`` is True, the returned value will be a
        string containing the tag name.

        If ``single`` is False (default), the returned value will be a list of
        strings containing tag names.
        """
        value = super(BaseTagField, self).clean(value)

        if self.tag_options.force_lowercase:
            value = value.lower()

        if single:
            return value

        try:
            return parse_tags(
                value, self.tag_options.max_count, self.tag_options.space_delimiter
            )
        except ValueError as e:
            raise forms.ValidationError(_("%s" % e))


class SingleTagField(BaseTagField):
    """
    Single tag field

    For parsing purposes, a single tag field wraps the contents in quotes
    """

    def prepare_value(self, value):
        """
        Prepare the value by stripping quotes
        """
        # Might have been passed a Tag object via initial - convert to string
        if isinstance(value, BaseTagModel):
            value = value.name
        tag_string = value
        return super(SingleTagField, self).prepare_value(tag_string)

    def _prepare_tag_options(self, tag_options):
        """
        Force max_count to 1
        """
        return tag_options + options.TagOptions(max_count=1)

    def widget_attrs(self, widget):
        """
        Given a Widget instance (*not* a Widget class), returns a dictionary of
        any HTML attributes that should be added to the Widget, based on this
        Field.
        """
        attrs = super(SingleTagField, self).widget_attrs(widget)
        attrs.update({"data-tag-type": "single"})
        return attrs

    def clean(self, value):
        """
        Clean by parsing tag with quotes
        """
        tag = super(SingleTagField, self).clean(value, single=True)
        if tag:
            return tag
        else:
            return None


class TagField(BaseTagField):
    """
    Form tag field
    """

    def prepare_value(self, value):
        """
        Prepare the value
        """
        # Could receive iterable of tags from a form's initial object
        if isinstance(value, TagModelQuerySet) or (
            hasattr(value, "__getitem__")
            and all(isinstance(tag, BaseTagModel) for tag in value)
        ):
            value = render_tags(value)

        # Deal with it as normal
        return super(TagField, self).prepare_value(value)


# ##############################################################################
# ###### Inline tagged formset
# ##############################################################################


class TaggedInlineFormSet(forms.models.BaseInlineFormSet):
    """
    An inline formset base class which works when the model is a tagged model
    using a SingleTagField, and the parent model is a tag model. Only affects
    this usage, otherwise operates as a normal InlineFormSet.

    When adding an inline instance, Django's BaseInlineFormSet sets the parent
    model instance by fk id instead of fk value. This normally makes sense, but
    it means Tagulous doesn't detect the change of tag, and can't update the
    count. This formset updates the count after saving.

    When editing an inline instance, Django's BaseInlineFormSet switches the
    fk form field from SingleTagField to InlineForeignKeyField, which expects
    the fk's pk - but when constructing the initial form data, it gets it from
    the original field's value_from_object, in this case will be the tag name.
    This formset corrects this back to use the fk's pk.
    """

    def save(self, *args, **kwargs):
        """Saves and updates the tag count, if parent model is a tag model"""
        obj = super(TaggedInlineFormSet, self).save(*args, **kwargs)
        if isinstance(self.instance, BaseTagModel):
            self.instance.update_count()
        return obj

    def _construct_form(self, i, **kwargs):
        # Construct as normal
        form = super(TaggedInlineFormSet, self)._construct_form(i, **kwargs)

        # SingleTagField will have broken the pk - fix it by clearing and
        # letting BoundField.value() use the field's initial value
        if isinstance(self.instance, BaseTagModel) and self.fk.name in form.initial:
            del form.initial[self.fk.name]
        return form



# -*- coding: utf-8 -*-
"""
Tagulous test: Tag models

Modules tested:
    tagulous.models.models.BaseTagModel
    tagulous.models.models.TagModel
    tagulous.models.models.TagModelManager
    tagulous.models.models.TagModelQuerySet
"""
import unittest

from django.db import IntegrityError
from django.test import TestCase

import tagulous.settings as tagulous_settings
from tagulous import models as tag_models
from tagulous import utils as tag_utils
from tagulous.settings import SLUG_TRUNCATE_UNIQUE
from tests.lib import TagTestManager
from tests.tagulous_tests_app import models as test_models


class TagModelTest(TagTestManager, TestCase):
    """
    Test tag model basics
    """

    manage_models = [
        test_models.MixedTest,
        test_models.MixedRefTest,
        test_models.NonTagRefTest,
    ]

    def setUpExtra(self):
        self.tag_model = test_models.MixedTestTagModel
        self.model1 = test_models.MixedTest
        self.model2 = test_models.MixedRefTest
        self.model_nontag = test_models.NonTagRefTest

    def test_tags_equal_instance(self):
        "Test TagModel.__eq__ with instances"
        t1a = self.tag_model.objects.create(name="one")
        t1b = self.tag_model.objects.get(name="one")
        self.assertEqual(t1a, t1b)

    def test_tags_equal_string(self):
        "Test TagModel.__eq__ with instance and string"
        t1 = self.tag_model.objects.create(name="one")
        self.assertEqual(t1, "one")

    def test_tags_not_equal_instance(self):
        "Test TagModel.__ne__ with instances"
        t1 = self.tag_model.objects.create(name="one")
        t2 = self.tag_model.objects.create(name="two")
        self.assertNotEqual(t1, t2)

    def test_tags_not_equal_string(self):
        "Test TagModel.__eq__ with instance and string"
        t1 = self.tag_model.objects.create(name="one")
        self.assertNotEqual(t1, "two")

    def test_get_absolute_url_defined(self):
        "Test get_absolute_url when passed in field definition"
        t1 = self.tag_model.objects.create(name="one")
        self.assertEqual(t1.get_absolute_url(), "url for one")

    def test_get_absolute_url_not_defined(self):
        "Test get_absolute_url when passed in field definition"
        t1 = test_models.SimpleMixedTest.tags.tag_model.objects.create(name="one")
        with self.assertRaises(AttributeError) as cm:
            t1.get_absolute_url()
        self.assertEqual(
            str(cm.exception),
            "'Tagulous_SimpleMixedTest_tags' has no attribute 'get_absolute_url'",
        )

    def test_name_field_length(self):
        """
        Value is initialized in setup.py runtests() settings
        """
        self.assertEqual(
            self.tag_model._meta.get_field("name").max_length,
            tagulous_settings.NAME_MAX_LENGTH,
        )

    def test_slug_field_length(self):
        """
        Value is initialized in setup.py runtests() settings
        """
        self.assertEqual(
            self.tag_model._meta.get_field("slug").max_length,
            tagulous_settings.SLUG_MAX_LENGTH,
        )

    def test_empty_name_raises_integrity_error(self):
        with self.assertRaises(IntegrityError) as cm:
            self.tag_model.objects.create(name=None)
        err = str(cm.exception).lower()
        self.assertTrue("name" in err)
        self.assertTrue("null" in err)

    def assertRelatedExists(self, related_fields, match_model, field_name):
        """
        Look through the related fields and find the field which refers to the
        specified model; fail if it does not exist
        """
        match_field = match_model._meta.get_field(field_name)
        for related in related_fields:
            if related.related_model == match_model and related.field == match_field:
                return related
        self.fail("Expected related field not found")

    def test_get_related_fields(self):
        "Check the class method returns a list of related fields"
        related_fields = self.tag_model.get_related_fields()
        self.assertEqual(len(related_fields), 4)

        # FK comes before M2M, so first two will be the SingleTagFields
        self.assertRelatedExists(related_fields, self.model1, "singletag")
        self.assertRelatedExists(related_fields, self.model2, "singletag")

        # Now the TagFields
        self.assertRelatedExists(related_fields, self.model1, "tags")
        self.assertRelatedExists(related_fields, self.model2, "tags")

    def test_get_related_fields_standard(self):
        "Check the class method can also find standard relationships"
        related_fields = self.tag_model.get_related_fields(include_standard=True)
        self.assertEqual(len(related_fields), 6)

        # SingleTagFields/FKs
        self.assertRelatedExists(related_fields, self.model1, "singletag")
        self.assertRelatedExists(related_fields, self.model2, "singletag")
        self.assertRelatedExists(related_fields, self.model_nontag, "fk")

        # TagFields/M2Ms
        self.assertRelatedExists(related_fields, self.model1, "tags")
        self.assertRelatedExists(related_fields, self.model2, "tags")
        self.assertRelatedExists(related_fields, self.model_nontag, "mm")

    def test_get_related_objects(self):
        """
        Check the class method returns a list of related models, fields and
        instances
        """
        t1 = self.create(self.model1, name="Test 1", singletag="Mr", tags="blue")
        t2 = self.create(self.model1, name="Test 2", singletag="Mr", tags="green")
        t3 = self.create(self.model2, name="Test 3", singletag="Mrs", tags="green")
        t4 = self.create(self.model2, name="Test 4", singletag="Mr", tags="green")

        #
        # Check Mr
        #
        singletag1 = self.tag_model.objects.get(name="Mr")
        rel_st1 = singletag1.get_related_objects()
        self.assertEqual(len(rel_st1), 2)
        # Don't assume order - we could probably guess reliably, but
        # why risk it in tests
        # rel_st1 is list of [model, field, [instances]]
        if rel_st1[0][0] == self.model1:
            rel_st1_m1 = rel_st1[0]
            rel_st1_m2 = rel_st1[1]
        else:
            rel_st1_m1 = rel_st1[1]
            rel_st1_m2 = rel_st1[0]

        self.assertEqual(rel_st1_m1[0], self.model1)
        self.assertEqual(rel_st1_m1[1], self.model1._meta.get_field("singletag"))
        self.assertEqual(len(rel_st1_m1[2]), 2)
        self.assertEqual(rel_st1_m1[2][0], t1)
        self.assertEqual(rel_st1_m1[2][1], t2)

        self.assertEqual(rel_st1_m2[0], self.model2)
        self.assertEqual(rel_st1_m2[1], self.model2._meta.get_field("singletag"))
        self.assertEqual(len(rel_st1_m2[2]), 1)
        self.assertEqual(rel_st1_m2[2][0], t4)

        #
        # Check Mrs
        #
        singletag2 = self.tag_model.objects.get(name="Mrs")
        rel_st2 = singletag2.get_related_objects()
        self.assertEqual(len(rel_st2), 1)
        rel_st2_m1 = rel_st2[0]
        self.assertEqual(rel_st2_m1[0], self.model2)
        self.assertEqual(rel_st2_m1[1], self.model2._meta.get_field("singletag"))
        self.assertEqual(len(rel_st2_m1[2]), 1)
        self.assertEqual(rel_st2_m1[2][0], t3)

        #
        # Check blue
        #
        tags1 = self.tag_model.objects.get(name="blue")
        rel_t1 = tags1.get_related_objects()
        self.assertEqual(len(rel_t1), 1)
        rel_t1_m1 = rel_t1[0]
        self.assertEqual(rel_t1_m1[0], self.model1)
        self.assertEqual(rel_t1_m1[1], self.model1._meta.get_field("tags"))
        self.assertEqual(len(rel_t1_m1[2]), 1)
        self.assertEqual(rel_t1_m1[2][0], t1)

        #
        # Check green
        #
        tags2 = self.tag_model.objects.get(name="green")
        rel_t2 = tags2.get_related_objects()
        self.assertEqual(len(rel_t2), 2)
        if rel_t2[0][0] == self.model1:
            rel_t2_m1 = rel_t2[0]
            rel_t2_m2 = rel_t2[1]
        else:
            rel_t2_m1 = rel_t2[1]
            rel_t2_m2 = rel_t2[0]

        self.assertEqual(rel_t2_m1[0], self.model1)
        self.assertEqual(rel_t2_m1[1], self.model1._meta.get_field("tags"))
        self.assertEqual(len(rel_t2_m1[2]), 1)
        self.assertEqual(rel_t2_m1[2][0], t2)

        self.assertEqual(rel_t2_m2[0], self.model2)
        self.assertEqual(rel_t2_m2[1], self.model2._meta.get_field("tags"))
        self.assertEqual(len(rel_t2_m2[2]), 2)
        self.assertEqual(rel_t2_m2[2][0], t3)
        self.assertEqual(rel_t2_m2[2][1], t4)

    def test_get_related_objects_flat(self):
        "Check the class method returns a flat list of related instances"
        t1 = self.create(self.model1, name="Test 1", singletag="Mr", tags="blue")
        t2 = self.create(self.model1, name="Test 2", singletag="Mr", tags="green")
        t3 = self.create(self.model2, name="Test 3", singletag="Mrs", tags="green")
        t4 = self.create(self.model2, name="Test 4", singletag="Mr", tags="green")

        # Check Mr
        singletag1 = self.tag_model.objects.get(name="Mr")
        rel_st1 = singletag1.get_related_objects(flat=True)
        self.assertEqual(len(rel_st1), 3)
        rel_st1.sort(key=lambda tag: tag.name)
        self.assertEqual(rel_st1[0], t1)
        self.assertEqual(rel_st1[1], t2)
        self.assertEqual(rel_st1[2], t4)

        # Check Mrs
        singletag2 = self.tag_model.objects.get(name="Mrs")
        rel_st2 = singletag2.get_related_objects(flat=True)
        self.assertEqual(len(rel_st2), 1)
        self.assertEqual(rel_st2[0], t3)

        # Check blue
        tags1 = self.tag_model.objects.get(name="blue")
        rel_t1 = tags1.get_related_objects(flat=True)
        self.assertEqual(len(rel_t1), 1)
        self.assertEqual(rel_t1[0], t1)

        # Check green
        tags2 = self.tag_model.objects.get(name="green")
        rel_t2 = tags2.get_related_objects(flat=True)
        self.assertEqual(len(rel_t2), 3)
        rel_t2.sort(key=lambda tag: tag.name)
        self.assertEqual(rel_t2[0], t2)
        self.assertEqual(rel_t2[1], t3)
        self.assertEqual(rel_t2[2], t4)

    def test_get_related_objects_flat_distinct(self):
        "Check the class method returns a flat list of distinct related instances"
        t1 = self.create(self.model1, name="Test 1", singletag="Mr", tags="blue")
        t2 = self.create(self.model2, name="Test 2", singletag="blue", tags="blue")

        # Check blue non-distinct
        tags1 = self.tag_model.objects.get(name="blue")
        rel_t1 = tags1.get_related_objects(flat=True)
        self.assertEqual(len(rel_t1), 3)
        rel_t1.sort(key=lambda tag: tag.name)
        self.assertEqual(rel_t1[0], t1)
        self.assertEqual(rel_t1[1], t2)
        self.assertEqual(rel_t1[2], t2)

        # Check blue distinct
        rel_t2 = tags1.get_related_objects(flat=True, distinct=True)
        self.assertEqual(len(rel_t2), 2)
        rel_t2.sort(key=lambda tag: tag.name)
        self.assertEqual(rel_t2[0], t1)
        self.assertEqual(rel_t2[1], t2)

    def test_get_related_objects_flat_include_standard(self):
        """
        Check the class method returns a flat list of related instances,
        including standard relationships

        No need to test other options with include_standard, uses same code
        """
        t1 = self.create(self.model1, name="Test 1", singletag="Mr", tags="blue")
        singletag1 = self.tag_model.objects.get(name="Mr")
        tags1 = self.tag_model.objects.get(name="blue")
        t2 = self.create(self.model_nontag, name="Test 2", fk=singletag1, mm=[tags1])

        # Check Mr
        rel_st1 = singletag1.get_related_objects(flat=True, include_standard=True)
        self.assertEqual(len(rel_st1), 2)
        rel_st1.sort(key=lambda tag: tag.name)
        self.assertEqual(rel_st1[0], t1)
        self.assertEqual(rel_st1[1], t2)

        # Check blue
        rel_t1 = tags1.get_related_objects(flat=True, include_standard=True)
        self.assertEqual(len(rel_t1), 2)
        rel_t1.sort(key=lambda tag: tag.name)
        self.assertEqual(rel_t1[0], t1)
        self.assertEqual(rel_t1[1], t2)

    def test_increment(self):
        "Increment the tag count"
        tag1 = self.create(self.tag_model, name="blue")
        self.assertInstanceEqual(tag1, count=0)
        tag1.increment()
        self.assertInstanceEqual(tag1, count=1)

    def test_increment_db(self):
        "Increment the tag count using the DB value, not in-memory"
        tag1 = self.create(self.tag_model, name="blue")
        tag2 = self.tag_model.objects.get(pk=tag1.pk)
        self.assertEqual(tag1.count, 0)
        self.assertEqual(tag2.count, 0)

        tag1.increment()
        self.assertInstanceEqual(tag1, count=1)
        self.assertInstanceEqual(tag2, count=1)
        self.assertEqual(tag1.count, 1)
        self.assertEqual(tag2.count, 0)

        tag2.increment()
        self.assertInstanceEqual(tag1, count=2)
        self.assertInstanceEqual(tag2, count=2)
        self.assertEqual(tag1.count, 1)
        self.assertEqual(tag2.count, 2)

    def test_decrement(self):
        tag1 = self.create(self.tag_model, name="blue", count=2)
        self.assertTagModel(self.tag_model, {"blue": 2})
        tag1.decrement()
        self.assertTagModel(self.tag_model, {"blue": 1})

    def test_decrement_delete(self):
        tag1 = self.create(self.tag_model, name="blue", count=1)
        self.assertTagModel(self.tag_model, {"blue": 1})
        tag1.decrement()
        self.assertTagModel(self.tag_model, {})

    def test_decrement_delete_protected(self):
        tag1 = self.create(self.tag_model, name="blue", count=1, protected=True)
        self.assertTagModel(self.tag_model, {"blue": 1})
        tag1.decrement()
        self.assertTagModel(self.tag_model, {"blue": 0})

    def test_decrement_delete_hasrefs(self):
        """
        Check that when a tag's count hits 0, but still has non-tag field
        references, that it isn't deleted - don't want to cascade/break refs
        """
        # Create tags with false count
        tag1 = self.create(self.tag_model, name="blue", count=1)
        tag2 = self.create(self.tag_model, name="red", count=1)
        self.assertTagModel(self.tag_model, {"blue": 1, "red": 1})

        # Create object with conventional references to tags
        t1 = self.create(self.model_nontag, name="CharField", fk=tag1, mm=[tag2])
        self.assertInstanceEqual(t1, name="CharField", fk=tag1, mm=[tag2])

        # No change to count
        self.assertTagModel(self.tag_model, {"blue": 1, "red": 1})

        # Check get_related_objects knows about them
        self.assertEqual(
            len(tag1.get_related_objects(flat=True, include_standard=True)), 1
        )
        self.assertEqual(
            len(tag2.get_related_objects(flat=True, include_standard=True)), 1
        )

        # Now decrement counts to 0, but tags remain
        tag1.decrement()
        tag2.decrement()
        self.assertTagModel(self.tag_model, {"blue": 0, "red": 0})

    def test_update_count(self):
        "Purposely knock the count off and update it"
        t1 = self.create(self.model1, name="Test 1", tags="blue")
        tag1 = self.tag_model.objects.get(name="blue")
        self.assertTagModel(self.tag_model, {"blue": 1})
        tag1.count = 3
        tag1.save()
        self.assertTagModel(self.tag_model, {"blue": 3})
        tag1.update_count()
        self.assertTagModel(self.tag_model, {"blue": 1})
        t1.delete()
        self.assertTagModel(self.tag_model, {})

    def test_slug_set(self):
        "Check the slug field is set correctly"
        t1a = self.tag_model.objects.create(name="One and Two!")
        self.assertEqual(t1a.slug, "one-and-two")

    def test_slug_saved(self):
        "Check the slug field is saved correctly"
        t1a = self.tag_model.objects.create(name="One and Two!")
        t1b = self.tag_model.objects.get(name="One and Two!")
        self.assertEqual(t1a.slug, t1b.slug)
        self.assertEqual(t1b.slug, "one-and-two")

    def test_slug_clash(self):
        "Check slug field avoids clashes to remain unique"
        t1a = self.tag_model.objects.create(name="one and two")
        t2a = self.tag_model.objects.create(name="One and Two!")
        t3a = self.tag_model.objects.create(name="One and Two?")
        t4a = self.tag_model.objects.create(name="One and Two.")
        self.assertEqual(t1a.slug, "one-and-two")
        self.assertEqual(t2a.slug, "one-and-two_1")
        self.assertEqual(t3a.slug, "one-and-two_2")
        self.assertEqual(t4a.slug, "one-and-two_3")

        t1b = self.tag_model.objects.get(name="one and two")
        t2b = self.tag_model.objects.get(name="One and Two!")
        t3b = self.tag_model.objects.get(name="One and Two?")
        t4b = self.tag_model.objects.get(name="One and Two.")
        self.assertEqual(t1b.slug, "one-and-two")
        self.assertEqual(t2b.slug, "one-and-two_1")
        self.assertEqual(t3b.slug, "one-and-two_2")
        self.assertEqual(t4b.slug, "one-and-two_3")

    def test_tag_model_factory(self):
        "Check the tag model factory supports setting max lengths"
        TestModel = test_models.TagSlugShorterModel
        self.assertEqual(TestModel._meta.get_field("name").max_length, 20)
        self.assertEqual(TestModel._meta.get_field("slug").max_length, 10)

    def test_long_slug_truncates(self):
        "Check the slug field is truncated correctly"
        TestModel = test_models.TagSlugShorterModel
        name_length = TestModel._meta.get_field("name").max_length
        slug_length = TestModel._meta.get_field("slug").max_length
        self.assertLess(slug_length, name_length)
        t1a = TestModel.objects.create(name="x" * name_length)
        self.assertEqual(t1a.name, "x" * name_length)
        self.assertEqual(t1a.slug, "x" * slug_length)

    def test_long_slug_clash_truncates(self):
        "Check clashing slug fields are truncated correctly"
        TestModel = test_models.TagSlugShorterModel
        name_length = TestModel._meta.get_field("name").max_length
        slug_length = TestModel._meta.get_field("slug").max_length
        self.assertLess(slug_length, name_length)
        ln1 = "{}{}".format("x" * (name_length - 1), "1")
        ln2 = "{}{}".format("x" * (name_length - 1), "2")
        t1a = TestModel.objects.create(name=ln1)
        t2a = TestModel.objects.create(name=ln2)
        self.assertEqual(t1a.name, ln1)
        self.assertEqual(t1a.slug, "x" * slug_length)
        self.assertEqual(t2a.name, ln2)
        slug2 = "{}_{}".format("x" * (slug_length - SLUG_TRUNCATE_UNIQUE), "1")
        self.assertEqual(t2a.slug, slug2)


# ##############################################################################
# ###### Test TagMeta in tag model
# ##############################################################################


class TagMetaTest(TagTestManager, TestCase):
    """
    Test TagMeta class. Builds on tests in tests_options.
    """

    def test_sets_options(self):
        "Check TagMeta sets the options"
        opt = test_models.TagMetaAbstractModel.tag_options

        # Check local options
        cls_opt = opt.items(with_defaults=False)
        self.assertEqual(cls_opt["initial"], ["Adam", "Brian", "Chris"])
        self.assertEqual(cls_opt["force_lowercase"], True)
        self.assertEqual(cls_opt["max_count"], 5)
        self.assertTrue("case_sensitive" not in cls_opt)

        # Check default options
        self.assertEqual(opt.case_sensitive, False)

    def test_inheritance(self):
        "Check TagMeta can be inherited and overridden"
        opt_abstract = test_models.TagMetaAbstractModel.tag_options
        opt = test_models.TagMetaModel.tag_options

        # Check they're not shared instances
        self.assertNotEqual(id(opt_abstract), id(opt))
        self.assertNotEqual(id(opt), id(tag_models.models.BaseTagModel.tag_options))

        # Check local options
        cls_opt = opt.items(with_defaults=False)
        self.assertEqual(cls_opt["case_sensitive"], True)
        self.assertEqual(cls_opt["max_count"], 10)

        # Local options will also include inherited options
        self.assertEqual(opt.initial, ["Adam", "Brian", "Chris"])
        self.assertEqual(opt.force_lowercase, True)
        self.assertEqual(opt.max_count, 10)
        self.assertEqual(opt.case_sensitive, True)


# ##############################################################################
# ###### Test unicode in tag model
# ##############################################################################


class TagModelUnicodeTest(TagTestManager, TestCase):
    """
    Test unicode tags - with unicode slugs disabled, and forced to not use unidecode,
    even if available
    """

    manage_models = [test_models.MixedTest]

    def setUpExtra(self):
        # Disable unidecode support
        self.unidecode_status = tag_utils.unidecode
        tag_utils.unidecode = None

        self.model = test_models.MixedTest
        self.tag_model = test_models.MixedTestTagModel
        self.o1 = self.create(
            self.model, name="Test", singletag="男の子", tags="boy, niño, 男の子"
        )

    def tearDownExtra(self):
        tag_utils.unidecode = self.unidecode_status

    def test_setup(self):
        "Check setup created tags as expected"
        self.assertTagModel(self.tag_model, {"boy": 1, "niño": 1, "男の子": 2})

    # Check lookup

    def test_get_singletag_get(self):
        "Check unicode singletag name matches"
        t1 = self.model.objects.get(singletag="男の子")
        self.assertEqual(t1.pk, self.o1.pk)

    def test_get_tag_ascii(self):
        "Check unicode tag name matches when ascii"
        t1 = self.model.objects.get(tags="boy")
        self.assertEqual(t1.pk, self.o1.pk)

    def test_get_tag_extended_ascii(self):
        "Check unicode tag name matches when extended ascii"
        t1 = self.model.objects.get(tags="niño")
        self.assertEqual(t1.pk, self.o1.pk)

    def test_get_tag_japanese(self):
        "Check unicode tag name matches when above extended ascii"
        t1 = self.model.objects.get(tags="男の子")
        self.assertEqual(t1.pk, self.o1.pk)

    # Check render

    def test_singletag_render(self):
        "Check unicode singletag name renders"
        t1 = self.model.objects.get(name="Test")
        self.assertEqual(str(t1.singletag), "男の子")

    def test_tag_render(self):
        "Check unicode tag name renders"
        t1 = self.model.objects.get(name="Test")
        tags = list(t1.tags.all())
        self.assertEqual(str(tags[0]), "boy")
        self.assertEqual(str(tags[1]), "niño")
        self.assertEqual(str(tags[2]), "男の子")

    def test_tag_string_render(self):
        "Check unicode tags string renders"
        t1 = self.model.objects.get(name="Test")
        self.assertEqual(str(t1.tags), "boy, niño, 男の子")

    # Check slugs

    def test_slug_ascii(self):
        "Check ascii tag name slugified to ascii"
        t1 = self.tag_model.objects.get(name="boy")
        self.assertEqual(t1.slug, "boy")

    def test_slug_extended_ascii(self):
        "Check extended ascii tag name slugified to ascii"
        t1 = self.tag_model.objects.get(name="niño")
        self.assertEqual(t1.slug, "nino")

    def test_slug_japanese(self):
        "Check unicode tag name slugified to ascii"
        name = "男の子"
        t1 = self.tag_model.objects.get(name=name)
        self.assertEqual(t1.name, name)
        self.assertEqual(t1.slug, "_")


try:
    from unidecode import unidecode
except ImportError:
    unidecode = None


class TagModelFullUnicodeTest(TagTestManager, TestCase):
    """
    Test unicode tags, with unicode enabled

    This only affects slugs
    """

    manage_models = [test_models.MixedTest]

    def setUpExtra(self):
        # Disable unicode support
        self.unicode_status = tagulous_settings.SLUG_ALLOW_UNICODE
        tagulous_settings.SLUG_ALLOW_UNICODE = True

        self.model = test_models.MixedTest
        self.tag_model = test_models.MixedTestTagModel
        self.o1 = self.create(
            self.model, name="Test", singletag="男の子", tags="boy, niño, 男の子"
        )

    def tearDownExtra(self):
        tagulous_settings.SLUG_ALLOW_UNICODE = self.unicode_status

    def test_setup(self):
        "Check setup created tags as expected"
        self.assertTagModel(self.tag_model, {"boy": 1, "niño": 1, "男の子": 2})

    def test_slug_ascii(self):
        "Check ascii tag name slugified to ascii"
        t1 = self.tag_model.objects.get(name="boy")
        self.assertEqual(t1.slug, "boy")

    def test_slug_extended_ascii(self):
        "Check extended ascii tag name slugified to ascii"
        t1 = self.tag_model.objects.get(name="niño")
        self.assertEqual(t1.slug, "niño")

    def test_slug_japanese(self):
        "Check unicode tag name slugified to ascii"
        name = "男の子"
        t1 = self.tag_model.objects.get(name=name)
        self.assertEqual(t1.slug, name)


# ##############################################################################
# ###### Test tag merging
# ##############################################################################


class TagModelMergeTest(TagTestManager, TestCase):
    """
    Test tag merging
    """

    manage_models = [test_models.MixedTest, test_models.MixedRefTest]

    def test_merge_tags(self):
        tag_model = test_models.MixedTestTagModel

        # Set up database
        a1 = self.create(test_models.MixedTest, name="a1", singletag="one", tags="one")
        a2 = self.create(test_models.MixedTest, name="a2", singletag="two", tags="two")
        a3 = self.create(
            test_models.MixedTest, name="a3", singletag="three", tags="three"
        )

        b1 = self.create(
            test_models.MixedRefTest, name="b1", singletag="one", tags="one"
        )
        b2 = self.create(
            test_models.MixedRefTest, name="b2", singletag="two", tags="two"
        )
        b3 = self.create(
            test_models.MixedRefTest, name="b3", singletag="three", tags="three"
        )

        # Confirm it's correct
        self.assertTagModel(tag_model, {"one": 4, "two": 4, "three": 4})
        self.assertInstanceEqual(a1, singletag="one", tags="one")
        self.assertInstanceEqual(a2, singletag="two", tags="two")
        self.assertInstanceEqual(a3, singletag="three", tags="three")
        self.assertInstanceEqual(b1, singletag="one", tags="one")
        self.assertInstanceEqual(b2, singletag="two", tags="two")
        self.assertInstanceEqual(b3, singletag="three", tags="three")

        # Merge tags
        s1 = tag_model.objects.get(name="one")
        s1.merge_tags(tag_model.objects.filter(name__in=["one", "two", "three"]))

        # Check it's correct
        self.assertTagModel(tag_model, {"one": 12})
        self.assertInstanceEqual(a1, singletag="one", tags="one")
        self.assertInstanceEqual(a2, singletag="one", tags="one")
        self.assertInstanceEqual(a3, singletag="one", tags="one")
        self.assertInstanceEqual(b1, singletag="one", tags="one")
        self.assertInstanceEqual(b2, singletag="one", tags="one")
        self.assertInstanceEqual(b3, singletag="one", tags="one")

    def test_merge_multiple_tags(self):
        "Test merging a queryset of multiple tags"
        tag_model = test_models.MixedTestTagModel

        # Set up database
        t1 = self.create(test_models.MixedTest, name="Test 1", tags="blue, green, red")
        t2 = self.create(test_models.MixedTest, name="Test 2", tags="blue, green, red")

        # Confirm it's correct
        self.assertTagModel(tag_model, {"blue": 2, "green": 2, "red": 2})
        self.assertInstanceEqual(t1, tags="blue, green, red")
        self.assertInstanceEqual(t2, tags="blue, green, red")

        # Merge tags
        s1 = tag_model.objects.get(name="blue")
        s1.merge_tags(tag_model.objects.filter(name__in=["blue", "green", "red"]))

        # Confirm it's correct
        self.assertTagModel(tag_model, {"blue": 2})
        self.assertInstanceEqual(t1, tags="blue")
        self.assertInstanceEqual(t2, tags="blue")

    def test_merge_by_name(self):
        "Test merging a list of tag names, including tags which don't exist"
        tag_model = test_models.MixedTestTagModel

        # Set up database
        t1 = self.create(test_models.MixedTest, name="Test 1", tags="blue, green, red")
        t2 = self.create(test_models.MixedTest, name="Test 2", tags="blue, green, red")

        # Confirm it's correct
        self.assertTagModel(tag_model, {"blue": 2, "green": 2, "red": 2})
        self.assertInstanceEqual(t1, tags="blue, green, red")
        self.assertInstanceEqual(t2, tags="blue, green, red")

        # Merge tags
        s1 = tag_model.objects.get(name="blue")
        s1.merge_tags(["blue", "green", "red", "pink"])

        # Confirm it's correct
        self.assertTagModel(tag_model, {"blue": 2})
        self.assertInstanceEqual(t1, tags="blue")
        self.assertInstanceEqual(t2, tags="blue")

    def test_merge_by_obj_list(self):
        "Test merging a list of tag objects"
        tag_model = test_models.MixedTestTagModel
        t1 = self.create(test_models.MixedTest, name="Test 1", tags="blue, green, red")
        t2 = self.create(test_models.MixedTest, name="Test 2", tags="blue, green, red")

        # Merge tags
        s1 = tag_model.objects.get(name="blue")
        s1.merge_tags(list(tag_model.objects.all()))

        # Confirm it's correct
        self.assertTagModel(tag_model, {"blue": 2})
        self.assertInstanceEqual(t1, tags="blue")
        self.assertInstanceEqual(t2, tags="blue")

    def test_merge_by_tag_string(self):
        "Test merging a tag string, including tags which don't exist"
        tag_model = test_models.MixedTestTagModel
        t1 = self.create(test_models.MixedTest, name="Test 1", tags="blue, green, red")
        t2 = self.create(test_models.MixedTest, name="Test 2", tags="blue, green, red")

        # Merge tags
        s1 = tag_model.objects.get(name="blue")
        s1.merge_tags("blue, green, red, pink")

        # Confirm it's correct
        self.assertTagModel(tag_model, {"blue": 2})
        self.assertInstanceEqual(t1, tags="blue")
        self.assertInstanceEqual(t2, tags="blue")


# ##############################################################################
# ###### Test tag model manager and queryset
# ##############################################################################


class TagModelQuerySetTest(TagTestManager, TestCase):
    """
    Test tag model queryset and manager
    """

    manage_models = [test_models.TagFieldOptionsModel]

    def setUpExtra(self):
        self.model = test_models.TagFieldOptionsModel
        self.tag_model = self.model.initial_list.tag_model
        self.o1 = self.model.objects.create(name="Test 1", initial_list="David, Eric")
        self.o2 = self.model.objects.create(name="Test 2", initial_list="Eric, Frank")

    def test_setup(self):
        self.assertTagModel(
            self.model.initial_list,
            {"Adam": 0, "Brian": 0, "Chris": 0, "David": 1, "Eric": 2, "Frank": 1},
        )

    def test_initial(self):
        initial_only = self.tag_model.objects.initial()
        self.assertEqual(len(initial_only), 3)
        self.assertEqual(initial_only[0], "Adam")
        self.assertEqual(initial_only[1], "Brian")
        self.assertEqual(initial_only[2], "Chris")

    def test_filter_or_initial(self):
        filtered = self.tag_model.objects.filter_or_initial(
            tagfieldoptionsmodel__name="Test 1"
        )
        self.assertEqual(len(filtered), 5)
        self.assertEqual(filtered[0], "Adam")
        self.assertEqual(filtered[1], "Brian")
        self.assertEqual(filtered[2], "Chris")
        self.assertEqual(filtered[3], "David")
        self.assertEqual(filtered[4], "Eric")

    def test_weight_scale_up(self):
        "Test weight() scales up to max"
        # Scale them to 2+2n: 0=2, 1=4, 2=6
        weighted = self.tag_model.objects.weight(min=2, max=6)
        self.assertEqual(len(weighted), 6)
        self.assertEqual(weighted[0].name, "Adam")
        self.assertEqual(weighted[0].weight, 2)

        self.assertEqual(weighted[1], "Brian")
        self.assertEqual(weighted[1].weight, 2)

        self.assertEqual(weighted[2], "Chris")
        self.assertEqual(weighted[2].weight, 2)

        self.assertEqual(weighted[3], "David")
        self.assertEqual(weighted[3].weight, 4)

        self.assertEqual(weighted[4], "Eric")
        self.assertEqual(weighted[4].weight, 6)

        self.assertEqual(weighted[5], "Frank")
        self.assertEqual(weighted[5].weight, 4)

    def test_weight_scale_down(self):
        "Test weight() scales down to max"
        # Add some extras so we can scale them 0.5n+2
        # Weight them so 0=2, 1=2 (rounded down), 4=4, 8=6

        # Eric will be used 8 times total - 6 more
        for i in range(6):
            self.model.objects.create(name="Test 3.%d" % i, initial_list="Eric")

        # Frank will be used 4 times total - 3 more
        for i in range(3):
            self.model.objects.create(name="Test 4.%d" % i, initial_list="Frank")

        self.assertTagModel(
            self.model.initial_list,
            {"Adam": 0, "Brian": 0, "Chris": 0, "David": 1, "Eric": 8, "Frank": 4},
        )

        weighted = self.tag_model.objects.weight(min=2, max=6)
        self.assertEqual(len(weighted), 6)
        self.assertEqual(weighted[0].name, "Adam")
        self.assertEqual(weighted[0].weight, 2)

        self.assertEqual(weighted[1], "Brian")
        self.assertEqual(weighted[1].weight, 2)

        self.assertEqual(weighted[2], "Chris")
        self.assertEqual(weighted[2].weight, 2)

        self.assertEqual(weighted[3], "David")
        self.assertEqual(weighted[3].weight, 2)

        self.assertEqual(weighted[4], "Eric")
        self.assertEqual(weighted[4].weight, 6)

        self.assertEqual(weighted[5], "Frank")
        self.assertEqual(weighted[5].weight, 4)

    def test_weight_integer(self):
        "Test weight() is a whole number"

        weighted = self.tag_model.objects.weight(min=2.5, max=6)
        self.assertEqual(weighted[0].name, "Adam")
        self.assertEqual(weighted[0].weight, 2)

    def test_weight_no_tags(self):
        "Test weight() when there are no tags"
        self.tag_model.objects.all().delete()
        self.assertEqual(self.tag_model.objects.count(), 0)

        weighted = self.tag_model.objects.weight()
        self.assertEqual(list(weighted.values_list("weight")), [])

    def test_weight_tags__no_associated(self):
        "Test weight() when there are tags but none are associated"
        self.tag_model.objects.all().delete()
        self.assertEqual(self.tag_model.objects.count(), 0)
        tag = self.tag_model.objects.create(name="test")
        self.assertEqual(tag.count, 0)

        weighted = self.tag_model.objects.weight()
        self.assertEqual(len(weighted), 1)
        self.assertEqual(weighted[0], "test")
        self.assertEqual(weighted[0].weight, 1)

    def test_weight_zero_no_tags(self):
        "Test weight() when there are no tags and minimum weight is zero"
        self.tag_model.objects.all().delete()
        self.assertEqual(self.tag_model.objects.count(), 0)

        weighted = self.tag_model.objects.weight(0, 6)
        self.assertEqual(list(weighted.values_list("weight")), [])

    def test_weight_zero_initial_tags(self):
        "Test weight() when there are tags but they have no count"
        self.assertEqual(self.tag_model.objects.count(), 6)

        weighted = self.tag_model.objects.weight(0, 6)
        self.assertEqual(
            list(weighted.values_list("weight", flat=True)), [0, 0, 0, 3, 6, 3]
        )

    def test_to_string(self):
        "Check manager and queryset can be converted to a tag string"
        self.assertEqual(
            str(self.tag_model.objects), "Adam, Brian, Chris, David, Eric, Frank",
        )
        self.assertEqual(
            str(self.tag_model.objects.all()), "Adam, Brian, Chris, David, Eric, Frank",
        )
        self.assertEqual(str(self.tag_model.objects.initial()), "Adam, Brian, Chris")
        self.assertEqual(str(self.o1.initial_list), "David, Eric")


# ##############################################################################
# ###### Test tag model with custom to_base
# ##############################################################################


class TagModelToBaseTest(TestCase):
    """
    Test autogenerated tag model with to_base
    """

    manage_models = [test_models.TagFieldOptionsModel]

    def test_custom_base_used(self):
        tag_model = test_models.CustomTagBaseTest.singletag.tag_model
        self.assertTrue(issubclass(tag_model, test_models.CustomTagBase))
        self.assertTrue(issubclass(tag_model, tag_models.TagModel))
        self.assertTrue(tag_model.is_custom)



"""
Test models
"""
from django.db import models

import tagulous


# ##############################################################################
# ###### Models for testing TagModel
# ##############################################################################


class TagMetaAbstractModel(tagulous.models.TagModel):
    """
    An abstract tag model with TagMeta definition
    """

    class Meta:
        abstract = True

    class TagMeta:
        initial = "Adam, Brian, Chris"
        force_lowercase = True
        max_count = 5


class TagMetaModel(TagMetaAbstractModel):
    """
    A tag model which inherits from TagMetaAbstractModel, with new and changed
    TagMeta values
    """

    class TagMeta:
        max_count = 10
        case_sensitive = True


class TagSlugShorterModel(tagulous.models.BaseTagModel):
    """
    A tag model with a slug field shorter than the name
    """

    name = models.CharField(max_length=20, unique=True)
    slug = models.SlugField(max_length=10)

    # Other fields we're not interested in testing but need to have anyway
    count = models.IntegerField(default=0)
    protected = models.BooleanField(default=False)

    class Meta:
        ordering = ("name",)
        unique_together = (("slug",),)


class TagMetaUser(models.Model):
    """
    A tagged model which uses the TagMetaModel
    """

    name = models.CharField(blank=True, max_length=100)
    two = tagulous.models.TagField(TagMetaModel, blank=True)


# ##############################################################################
# ###### Models for testing SingleTagField
# ##############################################################################


class SingleTagFieldModel(models.Model):
    """
    For testing simple single tag fields
    """

    name = models.CharField(blank=True, max_length=100)
    title = tagulous.models.SingleTagField(blank=True, null=True)


class SingleTagFieldConcreteInheritanceModel(SingleTagFieldModel):
    """
    Test concrete inheritance of a SingleTagField
    """


class SingleTagFieldAbstractInheritanceBase(models.Model):
    """
    Abstract base class for SingleTagFieldAbstractInheritanceModel
    """

    name = models.CharField(blank=True, max_length=100)
    title = tagulous.models.SingleTagField(blank=True, null=True)

    class Meta:
        abstract = True


class SingleTagFieldAbstractInheritanceModel(SingleTagFieldAbstractInheritanceBase):
    """
    Test abstract inheritance of a SingleTagField
    """


class SingleTagFieldAbstractInheritanceSecondModel(
    SingleTagFieldAbstractInheritanceBase
):
    """
    Test second abstract inheritance of a SingleTagField
    """


class SingleTagFieldOptionalModel(models.Model):
    """
    Test optional single tag fields
    """

    name = models.CharField(blank=True, max_length=100)
    tag = tagulous.models.SingleTagField(blank=True, null=True)


class SingleTagFieldRequiredModel(models.Model):
    """
    Test required single tag fields
    """

    name = models.CharField(blank=True, max_length=100)
    tag = tagulous.models.SingleTagField(blank=False, null=False)


class SingleTagFieldMultipleModel(models.Model):
    """
    For testing multiple single tag fields
    """

    name = models.CharField(blank=True, max_length=100)
    tag1 = tagulous.models.SingleTagField(blank=False, null=False)
    tag2 = tagulous.models.SingleTagField(blank=False, null=False)
    tag3 = tagulous.models.SingleTagField(blank=False, null=False)


class SingleTagFieldOptionsModel(models.Model):
    """
    For testing model and form SingleTagField options
    """

    name = models.CharField(blank=True, max_length=100)
    initial_string = tagulous.models.SingleTagField(
        blank=True, null=True, initial="Mr, Mrs, Ms"
    )
    initial_list = tagulous.models.SingleTagField(
        blank=True, null=True, initial=["Mr", "Mrs", "Ms"], autocomplete_initial=True
    )
    protect_initial_true = tagulous.models.SingleTagField(
        blank=True, null=True, protect_initial=True, initial="Mr"
    )
    protect_initial_false = tagulous.models.SingleTagField(
        blank=True, null=True, protect_initial=False, initial="Mr"
    )
    protect_all_true = tagulous.models.SingleTagField(
        blank=True, null=True, protect_all=True
    )
    protect_all_false = tagulous.models.SingleTagField(
        blank=True, null=True, protect_all=False
    )
    case_sensitive_true = tagulous.models.SingleTagField(
        blank=True, null=True, case_sensitive=True, initial="Mr"
    )
    case_sensitive_false = tagulous.models.SingleTagField(
        blank=True, null=True, case_sensitive=False, initial="Mr"
    )
    force_lowercase_true = tagulous.models.SingleTagField(
        blank=True, null=True, force_lowercase=True
    )
    force_lowercase_false = tagulous.models.SingleTagField(
        blank=True, null=True, force_lowercase=False
    )
    # max_count doesn't apply to SingleTagField
    autocomplete_view = tagulous.models.SingleTagField(
        blank=True, null=True, autocomplete_view="tagulous_tests_app-null"
    )
    autocomplete_limit = tagulous.models.SingleTagField(
        blank=True,
        null=True,
        autocomplete_limit=3,
        # Limit only takes effect when there's a view
        autocomplete_view="tagulous_tests_app-null",
    )
    autocomplete_settings = tagulous.models.SingleTagField(
        blank=True,
        null=True,
        autocomplete_settings={"setting1": 1, "setting2": True, "setting3": "example"},
    )

    class Meta:
        # Must set a short verbose name - tagulous auto-generated model
        # verbose names will be too long otherwise
        verbose_name = "STFOM"
        ordering = ("name",)


# ##############################################################################
# ###### Models for testing TagField
# ##############################################################################


class TagFieldModel(models.Model):
    """
    For testing basic tags
    """

    name = models.CharField(blank=True, max_length=100)
    tags = tagulous.models.TagField()


class TagFieldConcreteInheritanceModel(TagFieldModel):
    """
    Test concrete inheritance of a TagField
    """


class TagFieldAbstractInheritanceBase(models.Model):
    """
    Abstract base class for AbstractTagFieldModel
    """

    name = models.CharField(blank=True, max_length=100)
    tags = tagulous.models.TagField()

    class Meta:
        abstract = True


class TagFieldAbstractInheritanceModel(TagFieldAbstractInheritanceBase):
    """
    Test abstract inheritance of a TagField
    """


class TagFieldAbstractInheritanceSecondModel(TagFieldAbstractInheritanceBase):
    """
    Test second abstract inheritance of a TagField
    """


class TagFieldOptionalModel(models.Model):
    """
    Test optional tag fields
    """

    name = models.CharField(blank=True, max_length=100)
    tag = tagulous.models.TagField(blank=True)


class TagFieldRequiredModel(models.Model):
    """
    Test required tag fields
    """

    name = models.CharField(blank=True, max_length=100)
    tag = tagulous.models.TagField(blank=False, null=False)


class TagFieldMultipleModel(models.Model):
    """
    For testing multiple tag fields
    """

    name = models.CharField(blank=True, max_length=100)
    tags1 = tagulous.models.TagField(blank=False, null=False)
    tags2 = tagulous.models.TagField(blank=False, null=False)
    tags3 = tagulous.models.TagField(blank=False, null=False)


class TagFieldOptionsModel(models.Model):
    """
    For testing model and form TagField options
    """

    name = models.CharField(blank=True, max_length=100)
    initial_string = tagulous.models.TagField(blank=True, initial="Adam, Brian, Chris")
    initial_list = tagulous.models.TagField(
        blank=True, initial=["Adam", "Brian", "Chris"], autocomplete_initial=True
    )
    protect_initial_true = tagulous.models.TagField(
        blank=True, protect_initial=True, initial="Adam"
    )
    protect_initial_false = tagulous.models.TagField(
        blank=True, protect_initial=False, initial="Adam"
    )
    protect_all_true = tagulous.models.TagField(blank=True, protect_all=True)
    protect_all_false = tagulous.models.TagField(blank=True, protect_all=False)
    case_sensitive_true = tagulous.models.TagField(
        blank=True, case_sensitive=True, initial="Adam"
    )
    case_sensitive_false = tagulous.models.TagField(
        blank=True, case_sensitive=False, initial="Adam"
    )
    force_lowercase_true = tagulous.models.TagField(blank=True, force_lowercase=True)
    force_lowercase_false = tagulous.models.TagField(blank=True, force_lowercase=False)
    # case_sensitive_true_force_lowercase_true - abbreviated to avoid problems
    # with databases that have field name length limits
    cs_true_fl_true = tagulous.models.TagField(
        blank=True,
        case_sensitive=False,
        force_lowercase=True,
        verbose_name_singular="case sensitive test",
    )
    max_count = tagulous.models.TagField(blank=True, max_count=3)
    autocomplete_view = tagulous.models.TagField(
        blank=True, autocomplete_view="tagulous_tests_app-unlimited"
    )
    autocomplete_limit = tagulous.models.TagField(
        blank=True,
        autocomplete_limit=3,
        # Limit only takes effect when there's a view
        autocomplete_view="tagulous_tests_app-limited",
    )
    autocomplete_settings = tagulous.models.TagField(
        blank=True,
        autocomplete_settings={"setting1": 1, "setting2": True, "setting3": "example"},
    )

    class Meta:
        # Set a short verbose name for tagulous auto-generated verbose name
        verbose_name = "TFOM"
        ordering = ("name",)


# ##############################################################################
# ###### Models for testing a mix of fields
# ##############################################################################


class SimpleMixedTest(models.Model):
    """
    For tests which need a SingleTagField and TagField
    """

    name = models.CharField(max_length=10)
    singletag = tagulous.models.SingleTagField(blank=True)
    tags = tagulous.models.TagField(blank=True)


class MixedTestTagModel(tagulous.models.TagModel):
    class TagMeta:
        def get_absolute_url(self):
            return "url for %s" % self


class MixedTest(models.Model):
    """
    For tests where it's useful for the SingleTagField and TagField to share
    a tag model
    """

    name = models.CharField(max_length=10)
    singletag = tagulous.models.SingleTagField(
        MixedTestTagModel, related_name="mixed_singletag", blank=True
    )
    tags = tagulous.models.TagField(
        MixedTestTagModel, related_name="mixed_tags", blank=True
    )

    class Meta:
        ordering = ("name",)


class MixedRefTest(models.Model):
    """
    Multiple models referencing tag tables
    """

    name = models.CharField(max_length=10)
    singletag = tagulous.models.SingleTagField(
        MixedTest.singletag.tag_model, related_name="mixed_ref_singletag", blank=True
    )
    tags = tagulous.models.TagField(
        MixedTest.tags.tag_model, related_name="mixed_ref_tags", blank=True
    )


class NonTagRefTest(models.Model):
    """
    ForeignKeys and ManyToManyFields directly referencing a tag model
    """

    name = models.CharField(max_length=10)
    fk = models.ForeignKey(
        MixedTest.singletag.tag_model,
        related_name="non_tag_fk",
        blank=True,
        on_delete=models.CASCADE,
    )
    mm = models.ManyToManyField(
        MixedTest.tags.tag_model, related_name="non_tag_mm", blank=True
    )


class MixedNonTagModel(tagulous.models.TagModel):
    pass


class MixedNonTagRefTest(models.Model):
    """
    Tag fields and conventional relationships referencing a tag model
    """

    name = models.CharField(max_length=10)
    singletag = tagulous.models.SingleTagField(
        MixedNonTagModel, blank=True, related_name="singletags"
    )
    tags = tagulous.models.TagField(MixedNonTagModel, blank=True, related_name="tags")
    fk = models.ForeignKey(
        MixedNonTagModel,
        blank=True,
        null=True,
        related_name="fk",
        on_delete=models.CASCADE,
    )
    mm = models.ManyToManyField(MixedNonTagModel, blank=True, related_name="mm")


class MixedOrderTest(models.Model):
    """
    For testing ordering of a SingleTagField and TagField when next to other
    M2M and non-M2M fields
    """

    char1 = models.CharField(blank=True, max_length=10)
    fk1 = models.ForeignKey(
        MixedTest, related_name="order_fk1", on_delete=models.CASCADE
    )
    char2 = models.CharField(blank=True, max_length=10)
    single1 = tagulous.models.SingleTagField()
    char3 = models.CharField(blank=True, max_length=10)
    m2m1 = models.ManyToManyField(MixedTest, related_name="order_m2m1")
    char4 = models.CharField(blank=True, max_length=10)
    multi1 = tagulous.models.TagField()
    char5 = models.CharField(blank=True, max_length=10)
    m2m2 = models.ManyToManyField(MixedTest, related_name="order_m2m2")
    char6 = models.CharField(blank=True, max_length=10)
    fk2 = models.ForeignKey(
        MixedTest, related_name="order_fk2", on_delete=models.CASCADE
    )
    char7 = models.CharField(blank=True, max_length=10)


class MixedStringTagModel(tagulous.models.TagModel):
    pass


class MixedStringTo(models.Model):
    """
    A tagged model with fields which refers to a tag model by string, rather
    than by class
    """

    name = models.CharField(max_length=10)
    singletag = tagulous.models.SingleTagField(
        "MixedStringTagModel", related_name="tag_meta_string_singletag", blank=True
    )
    tags = tagulous.models.TagField(
        "MixedStringTagModel", related_name="tag_meta_string_tags", blank=True
    )


class MixedSelfTo(tagulous.models.TagModel):
    """
    A tagged tag model, with tag fields which refers itself using 'self'
    """

    alternate = tagulous.models.SingleTagField("self", blank=True)
    related = tagulous.models.TagField("self", blank=True)

    class TagMeta:
        force_lowercase = True


class TreeTest(models.Model):
    """
    For testing tag trees
    """

    name = models.CharField(max_length=10)
    singletag = tagulous.models.SingleTagField(tree=True, blank=True)
    tags = tagulous.models.TagField(tree=True, blank=True)


class CustomTagTree(tagulous.models.TagTreeModel):
    """
    Custom tag tree model
    """

    pass


class CustomTreeTest(models.Model):
    """
    For testing custom tag trees
    """

    name = models.CharField(max_length=10)
    singletag = tagulous.models.SingleTagField(
        "CustomTagTree", blank=True, related_name="custom_singletag"
    )
    tags = tagulous.models.TagField(
        "CustomTagTree", blank=True, related_name="custom_tags"
    )


class ManyToOneTest(models.Model):
    """
    Add a reverse FK to MixedRefTest for serialization tests
    """

    name = models.CharField(max_length=10)
    mixed_ref_test = models.ForeignKey(
        MixedRefTest, related_name="many_to_one", on_delete=models.CASCADE
    )


class CustomTagBase(tagulous.models.TagModel):
    """
    For testing custom tag base models for auto-generated tag models
    """

    is_custom = True

    class Meta:
        abstract = True


class CustomTagBaseTest(models.Model):
    """
    Test custom tag base model
    """

    name = models.CharField(max_length=10)
    singletag = tagulous.models.SingleTagField(
        to_base=CustomTagBase, blank=True, related_name="custom_singletag"
    )



"""
Tag options
"""
from .. import constants
from ..utils import parse_tags, render_tags


class TagOptions(object):
    """
    Simple class container for tag options
    """

    def __init__(self, **kwargs):
        """
        Set up tag options using defaults, overridden by keyword arguments
        """
        self.update(kwargs)

    def update(self, options):
        """
        Update this TagOptions in place with options from a dict or TagOptions
        """
        # Ensure a dict
        if isinstance(options, TagOptions):
            options = options.items(with_defaults=False)

        for key, val in options.items():
            setattr(self, key, val)

        return self

    def set_missing(self, options):
        """
        Update this TagOptions in place with options from a dict or TagOptions,
        but only where the keys are not already set on this TagOptions

        Similar to opt1 = opt2 + op1, only changes in place
        """
        # Ensure a dict
        if isinstance(options, TagOptions):
            options = options.items(with_defaults=False)

        # Get currently set keys
        items = self.items(with_defaults=False)
        for key, val in options.items():
            if key not in items:
                setattr(self, key, val)

        return self

    def contribute_to_class(self, cls, name):
        """
        Add to class
        """
        setattr(cls, name, self)

    def __setattr__(self, name, value):
        """
        Only allow an option to be set if it's valid
        """
        if name == "initial":
            # Store as a list of strings, with the tag string available on
            # initial_string for migrations
            if value is None:
                self.__dict__["initial_string"] = ""
                self.__dict__["initial"] = []
            elif isinstance(value, str):
                self.__dict__["initial_string"] = value
                self.__dict__["initial"] = parse_tags(value)
            else:
                self.__dict__["initial_string"] = render_tags(value)
                self.__dict__["initial"] = value

        elif name in constants.OPTION_DEFAULTS:
            self.__dict__[name] = value
        else:
            raise AttributeError(name)

    def __getattr__(self, name):
        """
        Fall back to default options if it's not set

        Requests for locally-defined properties will be fielded directly by
        __dict__
        """
        if name == "initial_string":
            # If here, there is no initial set in the __dict__
            # There is nothing in defaults to fall back to either
            return ""
        if name not in constants.OPTION_DEFAULTS:
            raise AttributeError(name)
        return self.__dict__.get(name, constants.OPTION_DEFAULTS[name])

    def _get_items(self, with_defaults, keys):
        """
        Return a dict of options specified in keys, with defaults if required
        """
        if with_defaults:
            return dict(
                [
                    (name, self.__dict__.get(name, constants.OPTION_DEFAULTS[name]))
                    for name in keys
                ]
            )

        return dict(
            [(name, value) for name, value in self.__dict__.items() if name in keys]
        )

    def items(self, with_defaults=True):
        """
        Get a dict of all options

        If with_defaults is True, any missing options will be set to their
        defaults; if False, missing options will be omitted.
        """
        return self._get_items(with_defaults, constants.OPTION_DEFAULTS)

    def form_items(self, with_defaults=True):
        """
        Get a dict of all options in FORM_OPTIONS, suitable for rendering
        into the data-tag-options attribute of the field HTML.

        If with_defaults is True, any missing options will be set to their
        defaults; if False, missing options will be omitted.
        """
        return self._get_items(with_defaults, constants.FORM_OPTIONS)

    def clone(self):
        """
        Return a new TagOptions object with the options set on this object.
        """
        dct = self.items(with_defaults=False)
        return TagOptions(**dct)

    def __add__(self, options):
        """
        Return a new TagOptions object with the options set on this object,
        overridden by any on the second specified TagOptions object.
        """
        dct = self.items(with_defaults=False)
        dct.update(options.items(with_defaults=False))
        return TagOptions(**dct)
